import * as React from 'react';
import {ParameterFieldProps} from "../ParameterFieldProps";
import {observer, useDisposable, useLocalStore} from "mobx-react-lite";
import {
    Builder,
    BuilderProps,
    Config,
    FieldOrGroup,
    Fields,
    Operators,
    Query,
    TypedMap,
    Types,
    Utils as QbUtils,
    WidgetConfigForType
} from 'react-awesome-query-builder';
import BasicConfig from 'react-awesome-query-builder/lib/config/antd';
import {ReportTargetModel} from "../../../../stores/ReportTargetStore";
import {computedFn} from "mobx-utils";
import {t} from "@lingui/macro";
import {i18n, useModelContext, useReadonlyContext} from "@rx-suite/ui";
import {Typography} from "@material-ui/core";
import equal from 'deep-equal'
import {reaction} from "mobx";
import moment from "moment";
import {ReportComponentModel} from "../../../../stores/ReportComponentStore";
import Widgets from 'react-awesome-query-builder/lib/components/widgets/antd';
const {FieldCascader} = Widgets;

const extractFieldDef = computedFn(function getDef(def): FieldOrGroup {
    if (def.subdefinition) {
        return {
            label: def.name,
            tooltip: def.desc,
            type: '!struct',
            subfields: Object.entries(def.subdefinition).reduce((memo: Fields, [key, def]: [string, any]) => {
                memo[key] = extractFieldDef(def);
                return memo;
            }, {})
        }
    } else {
        let type;
        let defaultValue;
        let fieldSettings = {};
        switch (def.type) {
            case 'integer':
            case 'float':
            case 'decimal':
            case 'bigint':
                defaultValue = '';
                type = 'number';
                break;
            case 'date':
            case 'datetime':
            case 'time':
                defaultValue = moment().toISOString();
                type = def.type;
                fieldSettings = {
                    dateFormat: 'YYYY-MM-DD'
                }
                break;
            case 'boolean':
                type = def.type;
                defaultValue = true;
                break;
            default:
                defaultValue = '';
                type = 'text';
        }
        return {
            type: type,
            defaultValue: defaultValue,
            fieldSettings: fieldSettings,
            label: def.name,
            tooltip: def.desc
        }
    }
});


export const FilterParameter: React.FunctionComponent<ParameterFieldProps> = observer(({parameter}) => {
    const component = useModelContext() as ReportComponentModel;
    const readOnly = useReadonlyContext();
    const target = component.viewModel.report_target as ReportTargetModel;
    const operators = React.useMemo(() => Object.entries((target.ui_hints || {}).filter_operations || {}).reduce((bucket: Operators, [key, def]: [string, any]) => {
        bucket[key] = {
            label: def.name,
            reversedOp: def.inverse,
            jsonLogic: key
        };
        return bucket;
    }, {}), [target]);

    const config = React.useMemo(() => ({
        ...BasicConfig,
        settings: {
            ...BasicConfig.settings,
            // @ts-ignore
            renderField: (props) => <FieldCascader {...props} />,
            renderSize: 'large',
            valueSourcesInfo: {
                value: {
                    label: i18n._(t`Value`)
                },
                field: {
                    label: i18n._(t`Field`),
                    widget: "fieldcascader",
                }
            },
            showNot: false,
            immutableGroupsMode: readOnly,
            immutableValuesMode: readOnly,
            immutableOpsMode: readOnly,
            immutableFieldsMode: readOnly,
        },
        types: Object.entries(BasicConfig.types).reduce((memo: Types, [key, def]: [string, any]) => {
            const ops = (key === 'boolean') ? Object.keys(operators).filter(o => o === '==' || o === '!=') : Object.keys(operators);
            memo[key] = {
                ...def,
                defaultOperator: '==',
                operators: ops,
                widgets: Object.entries(def.widgets as TypedMap<WidgetConfigForType>).reduce((widgets: TypedMap<WidgetConfigForType>, [widgetKey, widget]: [string, WidgetConfigForType]) => {
                    widgets[widgetKey] = {
                        ...widget,
                        operators: ops
                    };
                    return widgets
                }, {})
            };
            return memo
        }, {}),
        operators: operators,
        fields: Object.entries(target.returns.properties).reduce((memo: Fields, [key, def]: [string, any]) => {
            memo[key] = extractFieldDef(def);
            return memo;
        }, {}),
        widgets: {
            ...BasicConfig.widgets,
            text: {
                ...BasicConfig.widgets.text,
                valuePlaceholder: i18n._(t`Enter value or leave blank`)
            }
        }
    } as Config), [target, operators]);

    const renderBuilder = React.useCallback((props: BuilderProps) => (
        <div className="query-builder-container">
            <div className="query-builder">
                <Builder {...props} />
            </div>
        </div>
    ), []);

    // we don't read and write directly to the filter parameter since React-Query-Builder needs to track a bunch of internal state that we don't want (we just want the generated JSON logic)
    // So we use this workingTree for all required state, and use a custom onChange to extract the important bits
    const treeState = useLocalStore((component) => ({
        workingTree: (component.viewModel.parameters.filter && Object.keys(component.viewModel.parameters.filter).length > 0) ? QbUtils.loadFromJsonLogic(component.viewModel.parameters.filter, config) : QbUtils.loadFromJsonLogic({
            and: []
        }, config),
        updateWorkingTree(newTree: any) {
            this.workingTree = newTree;
            const serialized = QbUtils.jsonLogicFormat(newTree, config);
            if (serialized && (!serialized.errors || (serialized.errors.length === 0)) && !equal(component.viewModel.parameters.filter, serialized.logic, {strict: true})) {
                component.nestedAssignment(`parameters.filter`, serialized.logic);
            }
        }
    }), component);

    // capture updates from the other direction, syncing filters to working tree
    useDisposable(() => reaction(() => component.viewModel.parameters.filter, () => {
        const serialized = QbUtils.jsonLogicFormat(treeState.workingTree, config);
        if (!equal(component.viewModel.parameters.filter, serialized.logic)) {
            treeState.updateWorkingTree((component.viewModel.parameters.filter && Object.keys(component.viewModel.parameters.filter).length > 0) ? QbUtils.loadFromJsonLogic(component.viewModel.parameters.filter, config) : QbUtils.loadFromJsonLogic({
                and: []
            }, config))
        }
    }));

    if (!target) {
        return null;
    }

    return (
        <React.Fragment>
            <Typography variant={'h6'}>{i18n._(t`Filter`)}</Typography>
            <Query
                {...config}
                value={treeState.workingTree}
                onChange={treeState.updateWorkingTree}
                renderBuilder={renderBuilder}
            />
        </React.Fragment>
    )
});