import * as React from 'react';
import {useCallback} from 'react';
import {observer} from "mobx-react-lite";
import {ReportComponentModel} from "../../../../../stores/ReportComponentStore";
import {ColumnParameter} from "../../../../../stores/Types";
import {Card, CardContent, CardHeader, IconButton} from "@material-ui/core";
import {Delete} from "@material-ui/icons";
import {i18n, InputField, ReadonlyConsumer, SelectField, useModelContext} from "@rx-suite/ui";
import {makeStyles} from "@material-ui/core/styles";
import {t} from "@lingui/macro";
import {ColumnAggregationOptions} from "./ColumnAggregationOptions";

const useStyles = makeStyles(theme => ({
    root: {
        height: '100%'
    }
}));

export const ColumnItem: React.FunctionComponent<{ item: ColumnParameter }> = observer(({item}) => {
    const component = useModelContext() as ReportComponentModel;
    const classes = useStyles();
    const columnInfo = component.getReturnDef(item.accessor);
    const columnComputationOptions = component.aggregationOperationsForColumn(item.accessor, !component.hasGroupingColumn);
    const columnGroupingOptions = component.groupingOperationsForColumn(item.accessor);
    const handleDeleteClick = useCallback(() => {
        component.removeColumn(item);
    }, [component, item]);
    const handleAggregationChange = useCallback((_, newVal) => {
        component.setColumnProp(item, 'aggregation', newVal);
    }, [item, component]);
    const handleRenameChange = useCallback((e) => {
        component.setColumnProp(item, 'rename', e.target.value);
    }, [item]);
    const handleSortPriorityChange = useCallback((e) => {
        component.setColumnProp(item, 'sort_priority', (e.target.value > 0) ? e.target.value : undefined);
    }, [item]);
    const handleSortDirectionChange = useCallback((_, newVal) => {
        component.setColumnProp(item, 'sort_desc', (newVal === 'desc'));
    }, [item]);

    if (!columnInfo) {
        return null;
    }
    return (
        <ReadonlyConsumer>
            {(readOnly: boolean) => (
                <Card className={classes.root}>
                    <CardHeader
                        action={
                            <IconButton aria-label="settings" onClick={handleDeleteClick} disabled={readOnly}>
                                <Delete/>
                            </IconButton>
                        }
                        title={columnInfo.name}
                        subheader={component.groupingNameForColumn(item.accessor || '')}
                        titleTypographyProps={{variant: 'h6', title: columnInfo.desc}}
                        subheaderTypographyProps={{variant: 'subtitle1'}}
                    />
                    <CardContent>
                        <InputField label={i18n._(t`Rename Column`)}
                                    value={item.rename || ''}
                                    disabled={readOnly}
                                    onChange={handleRenameChange}/>
                        <SelectField name={'aggregation'}
                                     value={item.aggregation || ''}
                                     label={i18n._(t`Operation`)}
                                     disabled={readOnly}
                                     options={
                                         [{label: i18n._(t`None`), value: ''},
                                             {
                                                 label: i18n._(t`Compute`),
                                                 options: columnComputationOptions.map((o: { name: any; id: any; }) => ({
                                                     label: o.name,
                                                     value: o.id
                                                 }))
                                             }, {
                                             label: i18n._(t`Group`),
                                             options: columnGroupingOptions.map((o: { name: any; id: any; }) => ({
                                                 label: o.name,
                                                 value: o.id
                                             }))
                                         }]
                                     }
                                     onChange={handleAggregationChange}
                                     clearable
                        />
                        <ColumnAggregationOptions item={item}/>
                        <InputField label={i18n._(t`Sort Priority`)}
                                    value={item.sort_priority || ''}
                                    type={'number'}
                                    disabled={readOnly}
                                    onChange={handleSortPriorityChange}/>
                        {
                            item.sort_priority && <SelectField name={'sort_desc'}
                                                               value={item.sort_desc ? 'desc' : 'asc'}
                                                               label={i18n._(t`Sort Direction`)}
                                                               disabled={readOnly}
                                                               options={
                                                                   [{
                                                                       label: i18n._(t`Descending`),
                                                                       value: 'desc'
                                                                   }, {
                                                                       label: i18n._(t`Ascending`),
                                                                       value: 'asc'
                                                                   }]
                                                               }
                                                               onChange={handleSortDirectionChange}/>
                        }
                    </CardContent>
                </Card>
            )}
        </ReadonlyConsumer>
    )
});