import * as React from 'react';
import {observer, Observer, useLocalStore} from "mobx-react-lite";
import ReactTable, {Resize} from "react-table";
import {t} from "@lingui/macro";
import {CheckboxField, i18n} from "@rx-suite/ui";
import {ColumnParameter} from "../../../stores/Types";
import {isArrayLike} from "mobx";
import {IconButton, Typography} from "@material-ui/core";
import {makeStyles} from "@material-ui/core/styles";
import {sprintf} from "sprintf-js";
import moment from "moment";
// @ts-ignore there is no typescript file for this
import useDimensions from 'react-use-dimensions';
import WidthText from "width-text";
import {RxAppModel, RxAppStore} from "@rx-suite/stores";
import {PaginatedReportComponent} from "../../../stores/PaginatedReportComponent";
import {LiveTableHeader} from './live_table/LiveTableHeader'
import {LiveTableFooter} from "./live_table/LiveTableFooter";
import {ExpandLess, ExpandMore} from "@material-ui/icons";
import classNames from 'classnames';

type ReportComponentLiveTableProps = {
    paginatedComponent: PaginatedReportComponent,
    tableProps?: React.ComponentProps<typeof ReactTable>,
    isSub?: boolean,
    leftMargin?: number,
    dense?: boolean,
    resizeInfo?: {
        autosize: boolean,
        handleResizingsChange: (newResizings: Resize[]) => void,
        handleAutosizeChanged: (e: React.ChangeEvent<HTMLInputElement>) => void,
        setAvailableWidth: (w: number) => void,
        performAutosize: () => void
    }
};

const useStyles = makeStyles(theme => ({
    tableHeader: {
        whiteSpace: 'pre-line!important' as any,
        wordWrap: 'break-word!important' as any,
    },
    printableTableHeader: {
        whiteSpace: 'pre-line!important' as any,
        wordWrap: 'break-word!important' as any,
        boxShadow: 'none!important'
    },
    tableFooter: {
        borderTop: 'groove'
    },
    tableFooterCell: {
        whiteSpace: 'pre-line!important' as any,
        wordWrap: 'break-word!important' as any,
        padding: '1px!important',
    },
    tableCell: {
        // textAlign: 'center'
    },
    printableTableCell: {
        whiteSpace: 'pre-line!important' as any,
        wordWrap: 'break-word!important' as any,
        padding: '1px!important',
    },
    titleBar: {
        borderRadius: theme.spacing(1, 1, 0, 0),
        padding: theme.spacing(1, 2),
        borderWidth: 1,
        borderStyle: 'solid',
        backgroundColor: theme.palette.grey[200],
        borderColor: theme.palette.grey[300],
        display: "flex"
    },
    spacer: {
        flexGrow: 1
    },
    printableTable: {
        display: 'table!important', // for repeated headers
        borderWidth: 0,
        borderLeftWidth: 1,
        borderRightWidth: 1,
        borderStyle: 'solid',
        borderColor: theme.palette.grey[300],
        boxSizing: 'border-box',
    },
    webTable: {
        display: 'block!important', // for sticky headers
        overflow: 'unset!important', // for sticky headers
    },
    printableBody: {
        display: 'block!important'
    },
    printableTRGroup: {
        display: 'block!important',
        pageBreakInside: 'avoid',
        breakInside: 'avoid',
        borderWidth: 0,
        borderBottomWidth: '1px!important',
        borderStyle: 'solid',
        borderColor: `${theme.palette.grey[300]}!important`,
    },
    ReactTable: {
        display: 'block!important',
    }
}));

const expanderWidth = 33
const cellBorderWidth = 1;
const tableBorderWidth = 2;
const cellPaddingWidth = 5;
const cellBufferWidth = 20;

const measureText = (text: any, font?: string) => {
    if (text) {
        return Math.ceil(WidthText(text.toString(), font) + (2 * cellBorderWidth) + (2 * cellPaddingWidth))
    } else {
        return 0;
    }
};

export const ReportComponentLiveTable: React.FunctionComponent<ReportComponentLiveTableProps> = observer(({
                                                                                                              paginatedComponent,
                                                                                                              tableProps,
                                                                                                              isSub,
                                                                                                              leftMargin,
                                                                                                              dense,
                                                                                                              resizeInfo
                                                                                                          }) => {
        const appUI = RxAppStore.selectedModel as RxAppModel;
        const printView = appUI && appUI.isPrintView;
        const params = (new URL(document.location.href)).searchParams;
        const classes = useStyles();
        const tableStore = useLocalStore(() => ({
            columnID(c: ColumnParameter) {
                return `${c.accessor}.${c.aggregation}`
            },
            columnHeader(c: ColumnParameter) {
                const columnInfo = paginatedComponent.component.getReturnDef(c.accessor);
                return c.rename || ((columnInfo || {name: ''}).name)
            }
        }));
        const hideControlbar = isSub || (printView && paginatedComponent.pageNumber != 0);

        const pageInfo = useLocalStore(() => ({
                drillTable(row: any) {
                    const drillValues = paginatedComponent.component.primeGroupingColumn && paginatedComponent.component.primeGroupingColumn.accessor ? {
                        [paginatedComponent.component.primeGroupingColumn.accessor]: paginatedComponent.rowDig(row.original, paginatedComponent.component.primeGroupingColumn)
                    } : {};
                    let margin = (paginatedComponent.resizings.find(r => r.id === tableStore.columnID(paginatedComponent.component.primeGroupingColumn)) || {value: 0}).value || 0;
                    if (paginatedComponent.component.groupingColumns.length <= 1) {
                        margin += expanderWidth;
                    }

                    return <ReportComponentLiveTable
                        paginatedComponent={paginatedComponent.drillPage(row.index.toString(), drillValues)}
                        tableProps={{minRows: 1}}
                        resizeInfo={resizeStore}
                        leftMargin={margin}
                        isSub/>
                },
                formatValue(c: ColumnParameter, val: any, isFooter: boolean) {
                    try {
                        if(c.aggregation == 'group_with_column'){
                            return (val || '').split('&|>').join(' ');
                        }
                        if ([null, undefined, ''].includes(val) || Number.isNaN(val)) {
                            if (!isFooter && paginatedComponent.component.primeGroupingColumn && (paginatedComponent.component.primeGroupingColumn.accessor === c.accessor)) {
                                return i18n._(t`[Blank]`)
                            } else {
                                return null;
                            }
                        } else {
                            if (isArrayLike(val)) {
                                return val.map((vSub, i) => <div
                                    key={`${vSub}.${i}`}>{this.formatValue(c, vSub, isFooter)}</div>);
                            } else {
                                const colDef = paginatedComponent.component.getReturnDef(c.accessor);
                                if (colDef) {
                                    switch (colDef.type) {
                                        case 'integer':
                                        case 'decimal':
                                        case 'float':
                                        case 'bigint':
                                        case 'numeric':
                                            if (colDef.format) {
                                                const zeroPadBuffer = ((val < 0) || colDef.format.explicit_plus) ? 1 : 0; // we need to allocate one char for the sign
                                                const formatString = [
                                                    '%', // identifies this as a format string
                                                    colDef.format.explicit_plus && '+',
                                                    (colDef.format.zero_pad > 0) ? `0${colDef.format.zero_pad + zeroPadBuffer}` : '',
                                                    `.${colDef.format.decimals || 0}`,
                                                    'f' // identifies this as a number that may or may not have decimals
                                                ].filter(s => s).join('');
                                                return sprintf(formatString, val);
                                            }
                                        case 'date':
                                            if (colDef.format && val) {
                                                const parsed = moment(val);
                                                if (parsed.isValid()) {
                                                    return parsed.format('YYYY-MM-DD')
                                                }
                                            }
                                            break;
                                        case 'datetime':
                                            if (colDef.format && val) {
                                                const parsed = moment(val);
                                                if (parsed.isValid()) {
                                                    return parsed.format('YYYY-MM-DD HH:mm:ss')
                                                }
                                            }
                                            break;
                                        case 'boolean':
                                            return val ? '✓' : 'X'
                                        default:
                                            break;
                                    }
                                }
                                return val.toString();
                            }
                        }
                    } catch (err) {
                        return val.toString();
                    }
                },
                get reactTableColumns() {
                    const appUI = RxAppStore.selectedModel as RxAppModel;
                    const printView = appUI && appUI.isPrintView;
                    const cols = (printView || !paginatedComponent.component.hasGroupingColumn) ? [] : [{
                        Header: <IconButton
                            size={'small'}
                            onClick={paginatedComponent.fullyExpanded ? paginatedComponent.collapseAll : paginatedComponent.expandAll}>
                            {paginatedComponent.fullyExpanded ? <ExpandLess/> : <ExpandMore/>}
                        </IconButton>,
                        id: '_expanderCol',
                        expander: true,
                        width: expanderWidth
                    }];
                    return cols.concat((paginatedComponent.component.viewModel.parameters.columns || []).map((c: ColumnParameter) => {
                        return {
                            Header: tableStore.columnHeader(c),
                            accessor: (row: any) => {
                                return paginatedComponent.rowDig(row, c)
                            },
                            Cell: (cellInfo: { value: any }) => this.formatValue(c, cellInfo.value, false),
                            Footer: () => {
                                if (!paginatedComponent.totals || !paginatedComponent.totalsMappings || !c.aggregation) {
                                    return null;
                                }
                                let mapping = paginatedComponent.totalsMappings[c.accessor || ''] || '';
                                if (mapping[c.aggregation]) {
                                    mapping = mapping[c.aggregation]
                                }
                                const val = this.formatValue(c, ((isArrayLike(paginatedComponent.totals) ? paginatedComponent.totals[0] : paginatedComponent.totals) || {})[mapping], true);
                                const aggregationInfo = paginatedComponent.component.aggregationDetails(c.aggregation);
                                return <div title={tableStore.columnHeader(c)}>
                                    <div>{((val !== null) && aggregationInfo && aggregationInfo.name) ? aggregationInfo.name : ''}</div>
                                    <div>{this.formatValue(c, val, true)}</div>
                                </div>;
                            },
                            id: tableStore.columnID(c)
                        }
                    }));
                }
            })
        );

        const [measureRef, {width}] = useDimensions();
        const resizeStore = resizeInfo || useLocalStore(() => ({
            autosize: !printView || !localStorage.getItem('componentResizings'),
            availableWidth: undefined as number | undefined,
            setAvailableWidth(newWidth: number) {
                this.availableWidth = newWidth - ((!printView && paginatedComponent.component.hasGroupingColumn) ? (expanderWidth + (2 * cellBorderWidth)) : 0) - (2 * tableBorderWidth);
            },
            //this is the one we trigger when the user manually resizes
            handleResizingsChange(newResizings: Resize[]) {
                paginatedComponent.setResizings(newResizings);
                this.autosize = false;
            },
            handleAutosizeChanged(e: React.ChangeEvent<HTMLInputElement>) {
                this.autosize = e.target.checked
            },
            get lengthData() {
                const columns = (paginatedComponent.component.viewModel.parameters.columns || []) as ColumnParameter[];
                const lengthData = columns.map(c => {
                    const header = tableStore.columnHeader(c);
                    const splitHeader = header.split(" ") as string[];
                    const datum = {
                        id: tableStore.columnID(c),
                        longestCell: measureText(header, paginatedComponent.fontString) + cellBufferWidth,
                        longestWord: Math.max(...splitHeader.map(h => measureText(h, paginatedComponent.fontString))) + cellBufferWidth,
                        hasSpaces: (splitHeader.length > 1),
                        emptyColumn: true
                    }
                    const aggregationInfo = paginatedComponent.component.aggregationDetails(c.aggregation);
                    if (aggregationInfo && aggregationInfo.name) {
                        const name = aggregationInfo.name;
                        const nameSplit = name.split(" ") as string[];
                        datum.longestCell = Math.max(datum.longestCell, measureText(name, paginatedComponent.fontString));
                        datum.longestWord = Math.max(datum.longestWord, ...nameSplit.map(v => measureText(v, paginatedComponent.fontString)));
                        datum.hasSpaces = datum.hasSpaces || (nameSplit.length > 1);
                    }
                    return datum;
                }) as { id: string, longestCell: number, longestWord: number, hasSpaces: boolean, emptyColumn: boolean }[];
                (paginatedComponent.rows || []).forEach(row => {
                    columns.forEach((c, i) => {
                        const lengthDatum = lengthData[i]!;
                        let rowVals = paginatedComponent.rowDig(row, c);
                        if (rowVals) {
                            if (!isArrayLike(rowVals)) {
                                rowVals = [rowVals]
                            }
                            rowVals = rowVals.map((val: any) => (val || '').toString());
                            rowVals.forEach((rowVal: string) => {
                                const rowValSplit = rowVal.split(" ") as string[];
                                lengthDatum.longestCell = Math.max(lengthDatum.longestCell, measureText(rowVal, paginatedComponent.fontString));
                                lengthDatum.longestWord = Math.max(lengthDatum.longestWord, ...rowValSplit.map(v => measureText(v, paginatedComponent.fontString)));
                                lengthDatum.hasSpaces = lengthDatum.hasSpaces || (rowValSplit.length > 1);
                                lengthDatum.emptyColumn = false;
                            });
                        }
                    });
                });
                return lengthData;
            },
            performAutosize() {
                try {
                    if (this.availableWidth) {
                        let widths = new Map();
                        let totalSize = 0;
                        this.lengthData.forEach(data => {
                            widths.set(data.id, data.longestCell);
                            totalSize += data.longestCell;
                        });

                        let sizeDiff = this.availableWidth - totalSize;
                        if (sizeDiff > 0) {
                            //Distribute extra space evenly
                            let alloc = sizeDiff / widths.size;
                            widths.forEach((width, col) => widths.set(col, Math.floor(width + alloc)));
                        } else if (sizeDiff < 0) {
                            // try and shrink columns in a sensible manner
                            let availableDeallocations = new Map();
                            this.lengthData.forEach(data => {
                                let minWidth = data.longestWord;
                                let maxWidth = data.longestCell;
                                let wiggleRoom = maxWidth - minWidth;
                                if (wiggleRoom > 0) {
                                    availableDeallocations.set(data.id, wiggleRoom);
                                }
                            });
                            // favour shrinking columns with no content in them.  Then ones with no spaces in them.  Then prune equal-size portions until we get there
                            let forceBreak = false;
                            while ((sizeDiff < 0) && availableDeallocations.size && !forceBreak) {
                                let deallocationSet = new Map(Array.from(availableDeallocations).filter(([col]) => {
                                    const match = this.lengthData.find(d => d.id === col);
                                    if (match) {
                                        return match.emptyColumn;
                                    } else {
                                        return false;
                                    }
                                }));
                                if (deallocationSet.size === 0) {
                                    deallocationSet = new Map(Array.from(availableDeallocations).filter(([col]) => {
                                        const match = this.lengthData.find(d => d.id === col);
                                        if (match) {
                                            return match.hasSpaces;
                                        } else {
                                            return false;
                                        }
                                    }));
                                }
                                if (deallocationSet.size === 0) {
                                    deallocationSet = availableDeallocations;
                                }
                                let standardDeallocation = Math.floor(Array.from(deallocationSet.values()).reduce((a, b) => a + b, 0) / deallocationSet.size);
                                if (standardDeallocation <= 0) {
                                    forceBreak = true;
                                } else {
                                    deallocationSet.forEach((wiggleRoom, col) => {
                                        if (!forceBreak) {
                                            let thisDeallocation = Math.max(1, Math.round(Math.min(-sizeDiff, standardDeallocation, wiggleRoom)))
                                            widths.set(col, widths.get(col) - thisDeallocation);
                                            sizeDiff += thisDeallocation; //note that sizeDiff is negative here by definition - a positive size diff means we have space to allocate
                                            if (wiggleRoom === thisDeallocation) {
                                                availableDeallocations.delete(col);
                                            } else {
                                                availableDeallocations.set(col, wiggleRoom - thisDeallocation);
                                            }
                                            if (sizeDiff >= 0) {
                                                forceBreak = true;
                                            }
                                        }
                                    });
                                }
                            }
                        }
                        paginatedComponent.setResizings(Array.from(widths).map(([col, size]) => {
                            return {id: col, value: size}
                        }));
                        paginatedComponent.flagRanSizing();
                    }
                } catch (e) {
                    console.error(`Autosize failed: ${e.message}`)
                }
            }
        }));
        React.useEffect(() => {
            if (paginatedComponent.resizings.length === 0) {
                paginatedComponent.setResizings([])
            }
        }, [paginatedComponent.component.report_target]);
        React.useEffect(() => {
            if (resizeStore.autosize && !isSub && !paginatedComponent.loading) {
                window.dispatchEvent(new Event('resize'));
                resizeStore.setAvailableWidth(width);
                resizeStore.performAutosize();
            }
        }, [width, resizeStore.autosize, paginatedComponent.loading, paginatedComponent.fontSize]);

        const headerProps = React.useCallback(() => ({
            className: printView ? classes.printableTableHeader : classes.tableHeader
        }), [classes]);

        const footerProps = React.useCallback(() => ({
            paginatedComponent: paginatedComponent,
            needFooter: !isSub && (!printView || params.has('isLastPage')) && (!!paginatedComponent.totalsFetchError || (paginatedComponent.component.viewModel.parameters.columns || []).some((c: ColumnParameter) => !!c.aggregation))
        }), [classes]);

        const footerTrProps = React.useCallback(() => ({
            className: classes.tableFooter
        }), [classes]);

        const footerTDProps = React.useCallback(() => ({
            className: classes.tableFooterCell
        }), [classes]);

        const tdProps = React.useCallback(() => ({
            className: (printView || dense) ? classes.printableTableCell : classes.tableCell
        }), [classes]);

        const trGroupProps = React.useCallback(() => ({
            className: printView ? classes.printableTRGroup : ''
        }), [classes]);

        const tBodyProps = React.useCallback(() => ({
            className: printView ? classes.printableBody : ''
        }), [classes]);

        const getTableProps = React.useCallback(() => ({
            className: printView ? classes.printableTable : classes.webTable
        }), [classes]);


        if (paginatedComponent.mainFetchError) {
            return <div ref={measureRef}>
                <div>{i18n._(t`There was an error and the report could not be generated.  Please provide the following info to support:`)}</div>
                <div>{paginatedComponent.runLink}</div>
                <div>{moment().toString()}</div>
                <div>{paginatedComponent.mainFetchError}</div>
            </div>
        }

        return (<div style={{marginLeft: leftMargin || 0}} ref={measureRef}>
                {!hideControlbar && <div className={classes.titleBar}>
                    <Typography variant={'h6'}>{paginatedComponent.component.viewModel.name}</Typography>
                    <div className={classes.spacer}/>
                    {!printView && <CheckboxField label={i18n._(t`Autosize`)} checked={resizeStore.autosize}
                                                  onChange={resizeStore.handleAutosizeChanged}/>}
                </div>}
                <ReactTable
                    loading={paginatedComponent.loading}
                    data={paginatedComponent.rows || []}
                    onFetchData={paginatedComponent.getData}
                    manual

                    getTheadThProps={headerProps}
                    getTfootProps={footerProps}
                    getTfootTrProps={footerTrProps}
                    getTfootTdProps={footerTDProps}
                    getTdProps={tdProps}
                    getTableProps={getTableProps}
                    getTbodyProps={tBodyProps}
                    getTrGroupProps={trGroupProps}

                    TheadComponent={isSub ? () => null : LiveTableHeader}
                    TfootComponent={LiveTableFooter}
                    SubComponent={(!printView && paginatedComponent.component.hasGroupingColumn) ? pageInfo.drillTable : undefined}

                    className={classNames(classes.ReactTable, '-striped -highlight')}
                    collapseOnDataChange={false}
                    filterable={false}
                    columns={pageInfo.reactTableColumns}
                    sortable={!printView}
                    resizable={!printView}
                    sorted={paginatedComponent.reactTableSorts}
                    onSortedChange={paginatedComponent.setReactTableSorts}
                    resized={paginatedComponent.resizings}
                    onResizedChange={resizeStore.handleResizingsChange}
                    expanded={paginatedComponent.expansions}
                    onExpandedChange={paginatedComponent.setExpansions}

                    pages={paginatedComponent.pageCount}
                    page={paginatedComponent.pageNumber}
                    pageSize={paginatedComponent.pageSize}
                    onPageChange={paginatedComponent.setPageNumber}
                    onPageSizeChange={paginatedComponent.setPageSize}
                    showPaginationBottom={!printView}
                    minRows={printView ? 0 : undefined}

                    NoDataComponent={printView ? () => null : undefined}
                    previousText={i18n._(t`Previous`)}
                    nextText={i18n._(t`Next`)}
                    loadingText={i18n._(t`Loading...`)}
                    pageText={i18n._(t`Page`)}
                    ofText={i18n._(t`of`)}
                    rowsText={i18n._(t`items`)}
                    renderPageSizeOptions={({
                                                pageSize,
                                                pageSizeOptions,
                                                rowsSelectorText,
                                                onPageSizeChange,
                                                rowsText,
                                            }: {
                        pageSize: number,
                        pageSizeOptions: number[],
                        rowsSelectorText: string,
                        onPageSizeChange: (newPageSize: number) => void,
                        rowsText: string,
                    }) => <Observer>{() => <>
                        <span>{paginatedComponent.totalRecordCount} {i18n._(t`total rows`)}</span>
                        <span className="select-wrap -pageSizeOptions">
                                <select aria-label={rowsSelectorText}
                                        onChange={e => onPageSizeChange(Number(e.target.value))}
                                        value={pageSize}>
                                    {pageSizeOptions.map((option, i) => (
                                        // eslint-disable-next-line react/no-array-index-key
                                        <option key={i} value={option}>
                                            {`${option} ${rowsText}`}
                                        </option>
                                    ))}
                                </select>
                            </span>
                    </>}</Observer>}

                    {...(tableProps || {})}
                />
            </div>
        )
    })
;
