import * as React from 'react';
import {observer, useLocalStore} from "mobx-react-lite";
import {ButtonField, i18n, SplitButton} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import {
    makeExport,
    ReportComponentResultModel,
    ReportComponentResultStore
} from "../../../../stores/ReportComponentResultStore";
import {AssignmentReturned, Cached, GetApp} from "@material-ui/icons";
import {runInAction} from "mobx";
import {ReportComponentResultStatus} from "../../../report_component_results/ReportComponentResultStatus";
import {makeStyles} from "@material-ui/core/styles";
import {PaginatedReportComponent} from "../../../../stores/PaginatedReportComponent";

const useStyles = makeStyles(theme => ({
    resultWrapper: {
        position: 'relative',
        display: 'inline-block',
        margin: theme.spacing(1),
        width: `10em`
    }
}));

export const CSVDumpButton: React.FunctionComponent<{ paginatedComponent: PaginatedReportComponent }> = observer(({paginatedComponent}) => {
    const classes = useStyles();
    const resultStore = useLocalStore(() => ({
        resultID: undefined as number | undefined,
        get pending() {
            if (!this.resultID) {
                return false;
            }
            if (this.result) {
                return !!this.result.download_link;
            } else {
                return true;
            }
        },
        get result() {
            return ReportComponentResultStore.get(this.resultID) as ReportComponentResultModel;
        },
        run(level?: number) {
            if (level === -1) {
                makeExport(paginatedComponent, 'csv', {columns: paginatedComponent.component.columnsUngrouped}).then(resp => {
                    runInAction(() => {
                        this.resultID = resp.id;
                    })
                });
            } else if (level === undefined) {
                makeExport(paginatedComponent, 'csv').then(resp => {
                    runInAction(() => {
                        this.resultID = resp.id;
                    })
                });
            } else {
                makeExport(paginatedComponent, 'csv', {columns: paginatedComponent.component.groupedColumnsAtDepth(level)}).then(resp => {
                    runInAction(() => {
                        this.resultID = resp.id;
                    })
                });
            }
        },
        clearResult() {
            this.resultID = undefined
        }
    }));
    if (!paginatedComponent.component.supportsCSVDump) {
        return null;
    }

    // case: report is generating or generated
    if (resultStore.result) {
        const GoBackButton = <ButtonField text={i18n._(t`New CSV`)} onClick={resultStore.clearResult}
                                          startIcon={<Cached/>}/>
        if (resultStore.result.download_link) {
            // report done
            return <>
                {GoBackButton}
                <ButtonField text={i18n._(t`Download CSV`)} onClick={resultStore.result.download}
                             startIcon={<GetApp/>} download/>
            </>;
        } else {
            // report running
            return <>
                {GoBackButton}
                <div className={classes.resultWrapper}>
                    <div>{i18n._(t`CSV`)}</div>
                    <ReportComponentResultStatus result={resultStore.result as ReportComponentResultModel}/>
                </div>
            </>;
        }
    }
    // case: report has not been started
    if (paginatedComponent.component.hasGroupingColumn) {
        // we have options to pick
        const options = [...paginatedComponent.component.groupingColumns.map((c, i) => {
            const columnInfo = paginatedComponent.component.getReturnDef(c.accessor);
            const name = c.rename || ((columnInfo || {name: ''}).name);
            return {
                label: `${i18n._(t`Export CSV`)} (${name})`,
                handleClick: () => resultStore.run(i)
            }
        }), {
            label: `${i18n._(t`Export CSV`)} (${i18n._(t`Raw Data`)})`,
            handleClick: () => resultStore.run(-1)
        }];
        return <div style={{
            position: 'relative',
            display: 'inline-block'
        }}>
            <SplitButton options={options} groupProps={{color: 'primary', size: 'small', variant: 'contained'}}/>
        </div>;
    }
    // no options, just start
    return <ButtonField text={i18n._(t`Export (CSV)`)}
                        startIcon={<AssignmentReturned/>}
                        onClick={() => resultStore.run()}/>;
});
