import * as React from 'react';
import {observer} from "mobx-react-lite";
import {FormProps, i18n, ModelForm, ModelInput, ModelSelect, SelectField} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import {AutomaticReportModel} from "../../stores/AutomaticReportStore";
import {ReportComponentModel, ReportComponentStore} from "../../stores/ReportComponentStore";
import {ReportTargetModel} from "../../stores/ReportTargetStore";
import {Conditional, ModelCheckbox, ModelDateField, PDFTemplateSelect, UserPicker} from "@rx-suite/ui";
import {makeStyles} from "@material-ui/core/styles";
import {Toolbar, Typography} from '@material-ui/core';
import {MailToEmailsField} from "../../components/automatic_reports/MailToEmailsField";
import {moduleIsInstalled} from "@rx-suite/stores/build";
import {IfFalse} from "@rx-suite/ui/build/components/logic/IfFalse";
import {IfTrue} from "@rx-suite/ui/build/components/logic/IfTrue";
import moment from "moment";
import {AutomaticReportRunParams} from "../../components/automatic_reports/AutomaticReportRunParams";
import {ReportSelectField} from "../../components/automatic_reports/ReportSelectField";

const useStyles = makeStyles(theme => ({
    frequency: {
        display: 'flex'
    },
    period: {
        display: 'flex'
    },
    mailTo: {
        display: 'flex'
    },
    attachments: {
        display: 'flex'
    }
}));

export const AutomaticReportsForm: React.FunctionComponent<FormProps> = observer(({model, readOnly}) => {
    const classes = useStyles();
    React.useEffect(() => {
        if (!readOnly) {
            if (model.viewModel.report_component && !model.viewModel.name) {
                model.nestedAssignment('name', (model.viewModel.report_component as ReportComponentModel).name);
            }

            if (!model.viewModel.date_column) {
                const validOptions = Object.keys((model as AutomaticReportModel).supportedDateRangeColumns);
                if (!validOptions.find(o => o === model.viewModel.date_column)) {
                    const defaultDate = validOptions.find(o => o === ((model.report_target as ReportTargetModel || {}).ui_hints || {}).default_date) || validOptions[0];
                    model.nestedAssignment('date_column', defaultDate)
                }
            }
        }
    }, [model, model.viewModel.report_component_id]);

    return (
        <ModelForm model={model} readOnly={readOnly}>
            <ModelInput name={'name'}/>
            <ReportSelectField/>
            <Toolbar><Typography variant={'h5'}>{i18n._(t`Schedule`)}</Typography></Toolbar>
            <div className={classes.frequency}>
                <ModelInput name={'run_frequency_value'} type={'number'} inputProps={{min: 1}}/>
                <ModelSelect name={'run_frequency_unit'} options={(model as AutomaticReportModel).runFrequencyOptions}/>
            </div>
            {
                model.needsPeriod && <div className={classes.period}>
                    <ModelInput name={'period_value'} type={'number'} inputProps={{min: 1}}/>
                    <ModelSelect name={'period_unit'} options={(model as AutomaticReportModel).periodOptions}/>
                    <ModelSelect name={'date_column'}
                                 options={Object.entries((model as AutomaticReportModel).supportedDateRangeColumns).map(([key, c]: [string, any]) => ({
                                     value: key,
                                     label: c.name
                                 }))}/>
                </div>
            }
            <ModelDateField name={'start_date'}/>
            {
                model.supportsRunParams && <>
                    <Toolbar><Typography variant={'h5'}>{i18n._(t`Filter`)}</Typography></Toolbar>
                    <AutomaticReportRunParams report={model as AutomaticReportModel}/>
                </>
            }
            <Toolbar><Typography variant={'h5'}>{i18n._(t`Alert`)}</Typography></Toolbar>
            <div className={classes.mailTo}>
                <UserPicker name={'mail_to_users'} onChange={model.handleSelectChange}
                            label={model.label('mail_to_users')}
                            disabled={readOnly}
                            value={model.viewModel.mail_to_users}
                            hint={{
                                title: model.label('mail_to_users'),
                                text: model.attributeDescription('mail_to_users'),
                            }}
                            multiple/>
                <MailToEmailsField autoreport={model as AutomaticReportModel} readOnly={readOnly}/>
            </div>
            <div className={classes.attachments}>
                <ModelCheckbox name={'attach_link'}/>
                <ModelCheckbox name={'attach_csv'}/>
                {moduleIsInstalled('rx_pdf') && <div>
                    <ModelCheckbox name={'attach_pdf'}/>
                    {model.viewModel.attach_pdf &&
                    <PDFTemplateSelect name={'pdf_template_id'} onChange={model.handleSelectChange}
                                       disabled={readOnly}
                                       value={Number(model.viewModel.pdf_template_id)}/>}
                </div>}
            </div>
            {!model.isNew && <div>
                <Toolbar><Typography variant={'h5'}>{i18n._(t`Most Recent Results`)}</Typography></Toolbar>
                <Conditional condition={!!model.last_run_time}>
                    <IfTrue>
                        <Typography variant={'body1'}>{i18n._(t`Last Run Time`)}:&nbsp;
                            {moment(model.last_run_time as Date).format('YYYY-MM-DD HH:mm:ss')}
                        </Typography>
                        <Typography variant={'body1'}>{i18n._(t`Last Run Result`)}:&nbsp;
                            <Conditional condition={!!model.last_run_error}>
                                <IfTrue>{i18n._(t`Error`)}: {model.last_run_error}</IfTrue>
                                <IfFalse>{i18n._(t`Success`)}</IfFalse>
                            </Conditional>
                        </Typography>
                    </IfTrue>
                    <IfFalse><Typography
                        variant={'body1'}>{i18n._(t`Last Run Time: Never`)}</Typography></IfFalse>
                </Conditional>
                {model.computed_next_run && <Typography variant={'body1'}>{i18n._(t`Next Run Time`)}:&nbsp;
                    {moment(model.computed_next_run as Date).format('YYYY-MM-DD HH:mm:ss')}
                </Typography>}
            </div>}
        </ModelForm>
    )
});
