import * as React from 'react';
import {useState} from 'react';
import {observer} from "mobx-react-lite";
import {FormProps, PaddedPaper} from "@rx-suite/ui";
import {Tab, Tabs} from "@material-ui/core";
import {t} from "@lingui/macro";
import {i18n, TitleSetter} from "@rx-suite/ui/build";
import {ReportComponentBuilder} from "./ReportComponentBuilder";
import {ReportComponentRunner} from "./ReportComponentRunner";
import {ReportComponentModel} from "../../stores/ReportComponentStore";
import moment from "moment";
import {ReportTargetModel} from "../../stores/ReportTargetStore";
import {RxAppModel, RxAppStore} from "@rx-suite/stores";
import {PaginatedReportComponent} from "../../stores/PaginatedReportComponent";
import {reaction} from "mobx";

type ComponentFormProps = {
    model: ReportComponentModel
} & Omit<FormProps, 'model'>

export const ReportComponentForm: React.FunctionComponent<ComponentFormProps> = observer(({model, readOnly}) => {
    const params = (new URL(document.location.href)).searchParams;
    const appUI = RxAppStore.selectedModel as RxAppModel;
    const printView = appUI && appUI.isPrintView;
    const initialTab = printView ? 'runner' : (params.get('tab') || 'builder')
    const [tab, setTab] = useState(initialTab);
    const handleTabChange = (_: any, v: any) => {
        setTab(v);
    };

    const paginatedComponent = React.useMemo(() => {
        const wrapped = new PaginatedReportComponent(model);
        if (printView) {
            if (localStorage.getItem('componentOverride')) {
                try {
                    const parsed = JSON.parse(localStorage.getItem('componentOverride')!);
                    console.log('Component Override Loading');
                    console.dir(parsed);
                    model.updateFromJson(parsed);
                } catch (err) {
                    console.warn(`Failed to load pdf component override: ${err}, ${localStorage.getItem('componentOverride')}`)
                }
            }
            if (localStorage.getItem('componentResizings')) {
                try {
                    const parsed = JSON.parse(localStorage.getItem('componentResizings')!);
                    console.log('Column Resizings Loading');
                    console.dir(parsed);
                    wrapped.setResizings(parsed);
                } catch (err) {
                    console.warn(`Failed to load pdf column size override: ${err}, ${localStorage.getItem('componentResizings')}`)
                }
            }
            if (localStorage.getItem('componentFontSize') && Number(localStorage.getItem('componentFontSize'))) {
                const parsed = Number(localStorage.getItem('componentFontSize'));
                console.log(`Font Size Loading: ${parsed}`);
                wrapped.setFontSize(parsed);
            }
        }
        if (params.get('datetype')) {
            wrapped.setDateColumn(params.get('datetype')!);
        } else if (model.supportedDateRangeColumns && model.supportedDateRangeColumns) {
            wrapped.setDateColumn(model.supportedDateRangeColumns[0]!)
            const validOptions = Object.keys(model.supportedDateRangeColumns);
            const defaultDate = validOptions.find(o => o === ((model.viewModel.report_target as ReportTargetModel || {}).ui_hints || {}).default_date) || validOptions[0];
            if (defaultDate) {
                wrapped.setDateColumn(defaultDate);
            }
        }
        if (model.supportedRunFilters) {
            Object.keys(model.supportedRunFilters).forEach(param => {
                if (params.get(param)) {
                    wrapped.handleRunFilterChange(param, params.get(param)!.split("|"))
                }
            });
        }
        if (params.get('startdate') || params.get('enddate')) {
            wrapped.setStartDate(params.get('startdate') || moment().startOf('month'));
            wrapped.setEndDate(params.get('enddate') || moment().endOf('month'));
        } else {
            wrapped.setDateRange('this_month')
        }

        if (params.get('pageNumber')) {
            wrapped.setPageNumber(Number(params.get('pageNumber')));
        }
        if (params.get('pageSize')) {
            wrapped.setPageSize(Number(params.get('pageSize')));
        }

        if (params.get('datetype') || params.get('startdate') || params.get('enddate') || printView) {
            wrapped.commit();
        }

        return wrapped;
    }, [model]);
    React.useEffect(
        () => reaction(() => model.viewModel.report_target_id, paginatedComponent.clearRunFilters),
        [model]
    );

    if (printView) {
        return <ReportComponentRunner paginatedComponent={paginatedComponent}/>;
    } else {
        return (
            <PaddedPaper>
                <TitleSetter title={model.isNew ? i18n._(t`Create Report Component`) : model.viewModel.name}/>
                {!printView && <Tabs value={tab} onChange={handleTabChange}>
                    <Tab label={i18n._(t`Design`)}
                         value={'builder'}
                    />
                    <Tab label={i18n._(t`Run`)}
                         value={'runner'}
                    />
                </Tabs>}
                {(tab === 'builder') &&
                <ReportComponentBuilder model={paginatedComponent.component} readOnly={readOnly}/>}
                {(tab === 'runner') &&
                <ReportComponentRunner paginatedComponent={paginatedComponent}/>}
            </PaddedPaper>
        );
    }
});
