import * as React from 'react';
import {observer} from "mobx-react-lite";
import {ReportComponentDateRangePicker} from "../../components/report_components/runner/ReportComponentDateRangePicker";
import {ReportComponentRunParams} from "../../components/report_components/runner/ReportComponentRunParams";
import {isArrayLike} from "mobx";
import {ReportComponentLiveTable} from "../../components/report_components/runner/ReportComponentLiveTable";
import {CopyToClipboardButton} from "../../components/report_components/runner/runner_buttons/CopyToClipboardButton";
import {LiveViewButton} from "../../components/report_components/runner/runner_buttons/LiveViewButton";
import {CSVDumpButton} from "../../components/report_components/runner/runner_buttons/CSVDumpButton";
import {moduleIsInstalled, RxAppModel, RxAppStore} from "@rx-suite/stores";
import {PaginatedReportComponent} from "../../stores/PaginatedReportComponent";
import {makeStyles} from "@material-ui/core/styles";
import {PDFRenderFlag} from "../../components/report_components/runner/PDFRenderFlag";
import {ReportTargetModel} from "../../stores/ReportTargetStore";
import {ButtonField, Conditional, i18n, InputField} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import {PDFMenu} from "../../components/report_components/runner/PDFMenu";
import {PictureAsPdf} from "@material-ui/icons";
import {IfTrue} from "@rx-suite/ui/build/components/logic/IfTrue";
import {IfFalse} from "@rx-suite/ui/build/components/logic/IfFalse";
import {PrintWrapper} from "../../components/report_components/runner/PrintWrapper";
import moment from "moment";

const formatValue = (v: any) => {
    if ([null, undefined, ''].includes(v) || Number.isNaN(v)) {
        return null;
    } else {
        if (isArrayLike(v)) {
            return v.map((vSub, i) => <div key={`${vSub}.${i}`}>{formatValue(vSub)}</div>)
        } else {
            return v.toString()
        }
    }
};

const useStyles = makeStyles(theme => ({
    buttons: {
        display: 'flex',
        alignItems: 'center'
    }
}));


export const ReportComponentRunner: React.FunctionComponent<{ paginatedComponent: PaginatedReportComponent }> = observer(({paginatedComponent}) => {
    const classes = useStyles();
    const appUI = RxAppStore.selectedModel as RxAppModel;
    const printView = !!(appUI && appUI.isPrintView);
    const [showPDFMenu, setShowPDF] = React.useState(false);
    const [template_id, setTemplateId] = React.useState('');
    const target = paginatedComponent.component.viewModel.report_target as ReportTargetModel;
    if (!target) {
        return <div>{i18n._(t`You must select a report target`)}</div>;
    }

    return (
        <React.Fragment>
            <Conditional condition={printView}>
                <IfFalse>
                    <ReportComponentDateRangePicker paginatedComponent={paginatedComponent}/>
                    <ReportComponentRunParams paginatedComponent={paginatedComponent}/>
                    <div className={classes.buttons}>
                        <LiveViewButton paginatedComponent={paginatedComponent}/>
                        <CopyToClipboardButton paginatedComponent={paginatedComponent}/>
                        <InputField label={i18n._(t`Font Size (pt)`)} value={paginatedComponent.fontSize}
                                    type={'number'}
                                    inputProps={{min: 1}}
                                    onChange={(e) => {
                                        paginatedComponent.setFontSize(Number(e.target.value))
                                    }}/>
                        <Conditional condition={moduleIsInstalled('rx_pdf')}>
                            <IfTrue>
                                <ButtonField text={i18n._(t`Setup PDF`)}
                                             onClick={() => setShowPDF(!showPDFMenu)}
                                             startIcon={<PictureAsPdf/>}/>
                            </IfTrue>
                        </Conditional>
                        <CSVDumpButton paginatedComponent={paginatedComponent}/>
                    </div>
                    <PDFMenu paginatedComponent={paginatedComponent} template_id={template_id}
                             onTemplateIdChange={setTemplateId} open={showPDFMenu}/>
                </IfFalse>
            </Conditional>
            <Conditional condition={paginatedComponent.viewIndex > 0}>
                <IfTrue>
                    <PrintWrapper fontSize={paginatedComponent.fontSize} template_id={template_id}
                                  applyPaper={showPDFMenu && !printView}>
                        <ReportComponentLiveTable paginatedComponent={paginatedComponent}
                                                  dense={showPDFMenu}
                                                  key={paginatedComponent.viewIndex}/>
                    </PrintWrapper>
                    <PDFRenderFlag paginatedComponent={paginatedComponent}/>
                </IfTrue>
            </Conditional>
            <Conditional condition={printView}>
                <IfTrue>
                    {i18n._(t`Generated on`)} {moment().format('YYYY-MM-DD')} {i18n._(t`Covering`)} {paginatedComponent.viewModel.startDate || ''} - {paginatedComponent.viewModel.endDate || ''}
                </IfTrue>
            </Conditional>
        </React.Fragment>
    )
});
