import {attribute, model, Model, Types} from '@rx-suite/data';
import {i18n} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import * as yup from "yup";
import {action, computed, isArrayLike} from "mobx";
import {ReportComponentModel, ReportComponentStore} from "./ReportComponentStore";
import {ReportTargetModel, ReportTargetStore} from "./ReportTargetStore";
import {computedFn} from "mobx-utils";

const testPeriod = (report_component_id: number | undefined, schema: any) => {
    return schema.test(
        'need-period-val',
        i18n._(t`You must give your report a period`),
        async (period: any) => {
            if (report_component_id) {
                const component = await ReportComponentStore.fetchOne(report_component_id.toString());
                if (component) {
                    const needDate = Object.keys(component.supportedDateRangeColumns || {}).length > 0;
                    if (needDate) {
                        return !!period;
                    }
                }
            }
            return true;
        })
}

@model({
    transportLayer: "/api/automatic_reports",
    permissionTarget: "AutomaticReport",
    routingInfo: {
        list: '/automatic_reports',
        create: '/automatic_reports/create',
        read: '/automatic_reports/:id/read',
        update: '/automatic_reports/:id/edit'
    }
})
export class AutomaticReportModel extends Model {
    types: Types = {
        mail_to_users: 'json',
        mail_to_emails: 'json',
        run_params: 'json',
        start_date: 'date',
        last_run_time: 'date',
        computed_next_run: 'date',
    };

    @attribute({
        label: i18n._(t`Name`),
        description: i18n._(t`Name`),
        validation: yup.string().required(i18n._(t`You must give your report a name`))
    })
    name: string = '';

    @attribute({
        label: i18n._(t`Run Every`),
        validation: yup.number().integer(i18n._(t`You must use a whole number`)).required(i18n._(t`You must give your report a run frequency`))
    })
    run_frequency_value: number = 1;

    @attribute({
        label: i18n._(t`Unit`),
        description: i18n._(t`Determines how frequently the report will run`),
        validation: yup.string().required(i18n._(t`You must give your report a run frequency`))
    })
    run_frequency_unit: 'day' | 'week' | 'month' | 'year' = 'week';

    @attribute({
        label: i18n._(t`Go Back`),
        validation: yup.number().nullable().integer(i18n._(t`You must use a whole number`)).transform(function (value, originalvalue) {
            return this.isType(value) ? value : null;
        }).when('report_component_id', testPeriod)
    })
    period_value: number | null = 1;

    @attribute({
        label: i18n._(t`Unit`),
        validation: yup.string().nullable().when('report_component_id', testPeriod)
    })
    period_unit: 'day' | 'week' | 'month' | 'year' | null = 'week';

    @attribute({
        label: i18n._(t`Date Type`),
        description: i18n._(t`Determines how much data will be in the report when it runs`),
        validation: yup.string().nullable().when('report_component_id', testPeriod)
    })
    date_column: string | null = '';

    @attribute({
        label: i18n._(t`Starting At`),
        description: i18n._(t`Determines when the report will run`),
        validation: yup.date().required(i18n._(t`You must give your report a start date`))
    })
    start_date: Date = new Date();

    @attribute({
        label: i18n._(t`Report To Run`),
        connects: 'ReportComponentModel',
        description: i18n._(t`The selected report will be run on the selected schedule.  You can only select reports you have permission to view.`),
        validation: yup.number().required(i18n._(t`You must set a report target`)),
        autoTable: false
    })
    report_component_id: number | undefined = undefined;

    @attribute({
        label: i18n._(t`Mail To Users`),
        description: i18n._(t`Email these users when the report is complete`),
    })
    mail_to_users: number[] = [];

    @attribute({
        label: i18n._(t`Mail To Emails`),
        description: i18n._(t`Email these addresses when the report is complete.  Note that links cannot be sent (since unregistered users lack the eprmission to view them)`),
        validation: yup.array().of(yup.string().email(i18n._(t`Incorrectly formatted email(s)`)))
    })
    mail_to_emails: string[] = [];

    @attribute({
        label: i18n._(t`Send Link`),
        description: i18n._(t`The email alert should include a link to view the report online`),
        validation: yup.boolean().nullable()
    })
    attach_link: boolean = true;

    @attribute({
        label: i18n._(t`Send CSV`),
        description: i18n._(t`The email alert should include an attached CSV`),
        validation: yup.boolean().nullable()
    })
    attach_csv: boolean = false;

    @attribute({
        label: i18n._(t`Send PDF`),
        description: i18n._(t`The email alert should include an attached PDF`),
        validation: yup.boolean().nullable()
    })
    attach_pdf: boolean = true;

    @attribute({
        label: i18n._(t`Last Run Error`),
        description: i18n._(t`The error encountered when the report was last run (if any)`),
    })
    last_run_error: string | null = null;

    @attribute({
        label: i18n._(t`PDF Template`),
        description: i18n._(t`The template to use when generating the PDF`),
        validation: yup.string().when('attach_pdf', {
            is: true,
            then: yup.string().required(i18n._(t`You must select a PDF template to attach a pdf`)),
            otherwise: yup.string().nullable(),
        })
    })
    pdf_template_id: string | null = null;

    @attribute({
        label: i18n._(t`Last Run Time`),
        description: i18n._(t`The last time the report ran`),
    })
    last_run_time: Date | null = null;

    @attribute({
        label: i18n._(t`Next Run Time`),
        description: i18n._(t`The next time the report is scheduled to run`),
    })
    computed_next_run: Date | null = null;

    @attribute({
        label: i18n._(t`Run Params`),
        description: i18n._(t`Extra parameters to feed into the report filters`),
    })
    run_params: { [key: string]: string[] } = {};

    @computed
    get isUserUpdatable() {
        const component = ReportComponentStore.get(this.viewModel.report_component_id) as ReportComponentModel | null;
        if (!component) {
            return true;
        }
        return component.runPermitted;
    }

    @computed get runFrequencyOptions() {
        const pluralFrequency = !!(this.viewModel.run_frequency_value && this.viewModel.run_frequency_value > 1);
        return [
            {label: pluralFrequency ? i18n._(t`Days`) : i18n._(t`Day`), value: 'day'},
            {label: pluralFrequency ? i18n._(t`Weeks`) : i18n._(t`Week`), value: 'week'},
            {label: pluralFrequency ? i18n._(t`Months`) : i18n._(t`Month`), value: 'month'},
            {label: pluralFrequency ? i18n._(t`Years`) : i18n._(t`Year`), value: 'year'},
        ];
    }

    @computed get periodOptions() {
        const pluralPeriod = !!(this.viewModel.period_value && this.viewModel.period_value > 1);
        return [
            {label: pluralPeriod ? i18n._(t`Days`) : i18n._(t`Day`), value: 'day'},
            {label: pluralPeriod ? i18n._(t`Weeks`) : i18n._(t`Week`), value: 'week'},
            {label: pluralPeriod ? i18n._(t`Months`) : i18n._(t`Month`), value: 'month'},
            {label: pluralPeriod ? i18n._(t`Years`) : i18n._(t`Year`), value: 'year'},
        ];
    }

    @computed get friendlyRunFrequency() {
        const unit = this.runFrequencyOptions.find(o => o.value === this.viewModel.run_frequency_unit);
        return [this.viewModel.run_frequency_value, unit && unit.label].filter(t => t).join(' ');
    }

    @computed get friendlyRunPeriod() {
        const unit = this.periodOptions.find(o => o.value === this.viewModel.period_unit);
        return [this.viewModel.period_value, unit && unit.label, this.dateColumnName && `(${this.dateColumnName})`].filter(t => t).join(' ');
    }

    @computed
    get emailErrors() {
        const errors = [] as string[];
        this.viewModel.mail_to_emails.forEach((str, i) => {
            const err = this.errors[`mail_to_emails[${i}]`]
            if (err) {
                errors.push(str);
            }
        });
        return errors;
    }

    @computed
    get emailErrorString() {
        if (this.emailErrors.length) {
            return `${i18n._(t`Incorrectly formatted emails`)}: ${this.emailErrors.join(', ')}`
        }
        return '';
    }

    @computed
    get reportTarget() {
        const component = ReportComponentStore.get(this.viewModel.report_component_id) as ReportComponentModel | null;
        if (component) {
            return ReportTargetStore.get(component.report_target_id);
        }
        return null;
    }

    @computed
    get supportedDateRangeColumns() {
        const component = ReportComponentStore.get(this.viewModel.report_component_id) as ReportComponentModel | null;
        if (component) {
            return component.supportedDateRangeColumns;
        }
        return {};
    }

    @computed
    get dateColumnOptions() {
        return Object.keys(this.supportedDateRangeColumns || {});
    }

    @computed get dateColumnName() {
        if (!this.viewModel.date_column) {
            return '';
        }
        const def = this.supportedDateRangeColumns[this.viewModel.date_column];
        if (!def) {
            return '';
        }
        return def.name || '';
    }

    @computed
    get needsPeriod() {
        return this.dateColumnOptions.length > 0;
    }

    @computed get supportedRunFilters() {
        const component = ReportComponentStore.get(this.viewModel.report_component_id) as ReportComponentModel | null;
        if (component) {
            return component.supportedRunFilters;
        }
        return {};
    }

    @computed get supportsRunParams() {
        return (Object.keys(this.supportedRunFilters).length > 0)
    }

    runFilterValue = computedFn(function getVal(this: AutomaticReportModel, columnName: string) {
        return this.viewModel.run_params[columnName];
    });

    @action.bound
    handleRunFilterChange(name: string | undefined, val: string | string[]) {
        if(!isArrayLike(val)){
            const existing = this.runFilterValue(name!) || [];
            val = [...existing, val];
        }
        this.viewModel.run_params = {...this.viewModel.run_params, [name!]: val};
    }

    @action.bound
    clearRunParams() {
        this.viewModel.run_params = {};
    }
}

export const AutomaticReportStore = AutomaticReportModel.store;