import {action, computed, isArrayLike, observable, runInAction} from "mobx";
import {t} from "@lingui/macro";
import {ReportComponentModel} from "./ReportComponentStore";
import {i18n} from "@rx-suite/ui";
import moment, {Moment} from "moment";
import {computedFn, createViewModel} from "mobx-utils";
import {IViewModel} from "mobx-utils/lib/create-view-model";
import {ColumnParameter} from "./Types";
import {Resize} from "react-table";

function sleep(ms: number) {
    return new Promise(resolve => setTimeout(resolve, ms));
}

export class PaginatedReportComponent {
    @observable pageNumber = 0;
    @observable pageSize = 20;
    @observable totalRecordCount = -1;
    @observable loading = false;
    @observable mainFetchError = null as string | null;
    @observable totalsFetchError = null as string | null;
    @observable rows = [] as any[];
    @observable totals = null as any;
    @observable rowMappings = {} as any;
    @observable totalsMappings = null as any;

    @observable component: ReportComponentModel;
    @observable viewModel: this & IViewModel<this>;
    @observable drillPages = {} as { [key: string]: PaginatedReportComponent };

    @observable runFilters = {} as { [key: string]: string[] };
    @observable dateColumn = null as string | null;
    @observable startDate = null as string | Date | Moment | null;
    @observable endDate = null as string | Date | Moment | null;
    @observable viewIndex = 0;
    @observable hasRunMain = false;
    @observable hasRunTotals = false;
    @observable uiRanSizing = false;
    @observable uiRenderedTotals = false;
    @observable fontSize = 14;
    @observable resizings = [] as Resize[];
    @observable expansions = {} as { [i: number]: {} };

    // this is not observable on purpose.  See reactTableSorts
    internalRTSorts = [] as {
        id: string,
        desc: boolean
    }[];

    @action.bound
    clearRunFilters() {
        this.viewModel.runFilters = {};
    }

    constructor(toWrap: ReportComponentModel) {
        this.component = toWrap;
        this.viewModel = createViewModel(this);
    }

    @computed get pageCount() {
        return Math.ceil((this.totalRecordCount || 0) / (this.pageSize || 1))
    }

    @computed
    get dateColumnInfo() {
        if (this.viewModel.dateColumn) {
            return this.component.supportedDateRangeColumns[this.viewModel.dateColumn]
        } else {
            return undefined;
        }
    }

    @computed
    get selectedDateUsesTime() {
        if (this.dateColumnInfo) {
            return this.dateColumnInfo.type === 'datetime' || this.dateColumnInfo.type === 'time'
        } else {
            return false;
        }
    }

    @computed
    get dateRanges() {
        return [
            {label: i18n._(t`Today`), value: 'today', start: moment().startOf('day'), end: moment().endOf('day')},
            {
                label: i18n._(t`Yesterday`),
                value: 'yesterday',
                start: moment().subtract(1, 'day').startOf('day'),
                end: moment().subtract(1, 'day').endOf('day')
            },
            {
                label: i18n._(t`This Week`),
                value: 'this_week',
                start: moment().startOf('week'),
                end: moment().endOf('week')
            },
            {
                label: i18n._(t`Last Week`),
                value: 'last_week',
                start: moment().subtract(1, 'week').startOf('week'),
                end: moment().subtract(1, 'week').endOf('week')
            },
            {
                label: i18n._(t`This Month`),
                value: 'this_month',
                start: moment().startOf('month'),
                end: moment().endOf('month')
            },
            {
                label: i18n._(t`Last Month`),
                value: 'last_month',
                start: moment().subtract(1, 'month').startOf('month'),
                end: moment().subtract(1, 'month').endOf('month')
            },
            {
                label: i18n._(t`This Year`),
                value: 'this_year',
                start: moment().startOf('year'),
                end: moment().endOf('year')
            },
            {
                label: i18n._(t`Last Year`),
                value: 'last_year',
                start: moment().subtract(1, 'year').startOf('year'),
                end: moment().subtract(1, 'year').endOf('year')
            },
            {label: i18n._(t`Custom`), value: 'custom'},
        ]
    }

    @computed
    get currentDateRangeValue() {
        const match = this.dateRanges.find(r => {
            const startSame = (!r.start || r.start.isSame(this.viewModel.startDate))
            const endSame = (!r.end || r.end.isSame(this.viewModel.endDate))
            return startSame && endSame;
        });
        return (match || {value: 'custom'}).value
    }

    @computed get changesPending() {
        return this.viewModel.isDirty
    }

    @computed
    get dateColumnOptions() {
        return Object.keys(this.component.supportedDateRangeColumns || {});
    }

    @computed
    get dateColumnError() {
        if (this.dateColumnOptions && (this.dateColumnOptions.length > 0) && (this.dateColumnOptions.find(c => c === this.viewModel.dateColumn))) {
            return i18n._(t`Date type is required`);
        } else {
            return undefined;
        }
    }

    @computed
    get startDateError() {
        if ((!this.viewModel.startDate || !moment(this.viewModel.startDate).isValid()) && this.dateColumnOptions && (this.dateColumnOptions.length > 0)) {
            return i18n._(t`Start date is not recognized`);
        } else {
            return undefined;
        }
    }

    @computed
    get endDateError() {
        if ((!this.viewModel.endDate || !moment(this.viewModel.endDate).isValid()) && this.dateColumnOptions && (this.dateColumnOptions.length > 0)) {
            return i18n._(t`End date is not recognized`);
        } else {
            return undefined;
        }
    }

    @computed
    get runLink() {
        const url = new URL(document.location.href);
        url.searchParams.set('tab', 'runner');
        url.searchParams.set('datetype', this.viewModel.dateColumn || '');
        if (this.selectedDateUsesTime) {
            url.searchParams.set('startdate', this.viewModel.startDate ? moment(this.viewModel.startDate).toISOString() : '');
            url.searchParams.set('enddate', this.viewModel.endDate ? moment(this.viewModel.endDate).toISOString() : '');
        } else {
            url.searchParams.set('startdate', this.viewModel.startDate ? moment(this.viewModel.startDate).format('YYYY-MM-DD') : '');
            url.searchParams.set('enddate', this.viewModel.endDate ? moment(this.viewModel.endDate).format('YYYY-MM-DD') : '');
        }
        url.searchParams.delete('forlink');
        url.searchParams.delete('backlink');
        Object.entries(this.viewModel.runFilters).forEach(([key, value]) => {
            url.searchParams.set(key, value.join("|"));
        });
        return url.toString();
    }

    @action.bound setPageNumber(newPageNumber: number) {
        this.pageNumber = newPageNumber;
    }

    @action.bound setPageSize(newPageSize: number) {
        this.pageSize = newPageSize;
    }

    // mostly for the PDF workflow
    async computeRecordCount() {
        return this.component.getData({
                skip_record_count: false,
                pagination: {limit: 0, offset: 0},
                columns: this.component.columnsWithSingleGroup,
                filter: this.component.filterWithRunParams(this.viewModel.dateColumn, this.viewModel.startDate, this.viewModel.endDate, this.viewModel.runFilters)
            }
        ).then((data: { total_records: number }) => {
            runInAction(() => {
                this.totalRecordCount = data.total_records;
            });
            return data.total_records
        });
    }

    getSourceUrls = async () => {
        return this.computeRecordCount().then(count => {
            const sources = [] as { url: string, render_selector?: string }[]
            const pageSize = 5000;
            const fetchLimit = Math.ceil((count / pageSize));
            if (fetchLimit === 0) {
                const url = new URL(this.runLink);
                url.searchParams.set('printView', 'true');
                url.searchParams.set('pageNumber', '0')
                url.searchParams.set('pageSize', '1')
                url.searchParams.set('isLastPage', 'true');
                sources.push({url: url.toString(), render_selector: '#reportRendered'})
            } else {
                for (let i = 0; i < fetchLimit; i++) {
                    const url = new URL(this.runLink);
                    url.searchParams.set('printView', 'true');
                    url.searchParams.set('pageNumber', i.toString())
                    url.searchParams.set('pageSize', pageSize.toString())
                    if (i === (fetchLimit - 1)) {
                        url.searchParams.set('isLastPage', 'true');
                    }
                    sources.push({url: url.toString(), render_selector: '#reportRendered'})
                }
            }
            return sources;
        })
    }

    @action.bound
    getData(state: { pageSize: number, page: number }, isRetry: boolean = false) {
        try {
            this.loading = true;
            this.mainFetchError = null;
            this.pageNumber = state.page;
            this.pageSize = state.pageSize;
            this.component.getData({
                    skip_record_count: this.totalRecordCount > 0,
                    pagination: {limit: state.pageSize, offset: state.page * state.pageSize},
                    columns: this.component.columnsWithSingleGroup,
                    filter: this.component.filterWithRunParams(this.viewModel.dateColumn, this.viewModel.startDate, this.viewModel.endDate, this.viewModel.runFilters)
                }
            ).then(async (data: { hits: any[], column_mappings: {}, total_records?: number | null, err?: string }) => {
                try {
                    runInAction(() => {
                        this.rows = data.hits || [];
                        this.rowMappings = data.column_mappings;
                        if (data.total_records || (data.total_records === 0)) {
                            this.totalRecordCount = data.total_records;
                        }
                        this.loading = false;
                        this.hasRunMain = true;
                    });
                } catch (err) {
                    if (isRetry) {
                        runInAction(() => {
                            this.mainFetchError = (data || {}).err || err.toString();
                            this.loading = false;
                            this.hasRunMain = true;
                        });
                    } else {
                        console.warn(`Failed to fetch ${err}`)
                        await sleep(2000);
                        this.getData(state, true)
                    }
                }
            });

            if (!this.totals) {
                try {
                    this.component.getTotals({filter: this.component.filterWithRunParams(this.viewModel.dateColumn, this.viewModel.startDate, this.viewModel.endDate, this.viewModel.runFilters)}).then((totals: any) => {
                            runInAction(() => {
                                if (totals !== undefined) {
                                    this.totalsFetchError = null;
                                    this.totals = (totals || {}).hits || {}; // could be intentionally null if no totals
                                    this.totalsMappings = (totals || {}).column_mappings || {};
                                    this.hasRunTotals = true;
                                } else {
                                    this.totalsFetchError = i18n._(t`"Failed to fetch overall totals, check server log`);
                                    this.totals = null;
                                    this.totalsMappings = null;
                                    this.hasRunTotals = true;
                                }
                            });
                        },
                        (err: any) => {
                            runInAction(() => {
                                this.totalsFetchError = err.toString();
                                this.totals = null;
                                this.totalsMappings = null;
                                this.hasRunTotals = true;
                            });
                        })

                } catch (err) {
                    this.totalsFetchError = err.toString();
                    this.totals = null;
                    this.totalsMappings = null;
                    this.hasRunTotals = true;
                }
            } else {
                this.hasRunTotals = true;
            }
        } catch (err) {
            this.mainFetchError = err.toString();
            this.loading = false;
            if (isRetry) {
                this.mainFetchError = err.toString();
                this.loading = false;
                this.hasRunMain = true;
            } else {
                console.warn(`failed to fetch ${err}`)
                sleep(2000).then(() => {
                    this.getData(state, true)
                });
            }
        }
    }

    @action.bound
    setDateRange(newRange: string) {
        const match = this.dateRanges.find(r => r.value === newRange);
        if (match) {
            this.viewModel.startDate = match.start || null;
            this.viewModel.endDate = match.end || null;
        }
    }

    @action.bound
    commit() {
        this.viewModel.submit();
        this.viewIndex += 1;
        this.totalRecordCount = -1;
        this.totals = null;
    }

    @action.bound
    cloneDates(other: PaginatedReportComponent) {
        this.setDateColumn(other.dateColumn);
        this.setStartDate(other.startDate);
        this.setEndDate(other.endDate);
    }

    @action.bound
    setDateColumn(newColumn: string | null) {
        this.dateColumn = newColumn;
    }

    @action.bound
    setStartDate(newDate: string | Date | Moment | null) {
        if (!newDate) {
            this.viewModel.startDate = null;
        } else if (typeof newDate === 'string') {
            this.viewModel.startDate = newDate;
        } else if (this.selectedDateUsesTime) {
            this.viewModel.startDate = moment(newDate).format('YYYY-MM-DD HH:mm:ss');
        } else {
            this.viewModel.startDate = moment(newDate).format('YYYY-MM-DD');
        }
    }

    @action.bound
    setEndDate(newDate: string | Date | Moment | null) {
        if (!newDate) {
            this.viewModel.endDate = null;
        } else if (typeof newDate === 'string') {
            this.viewModel.endDate = newDate;
        } else if (this.selectedDateUsesTime) {
            this.viewModel.endDate = moment(newDate).format('YYYY-MM-DD HH:mm:ss');
        } else {
            this.viewModel.endDate = moment(newDate).format('YYYY-MM-DD');
        }
    }

    rowDig(row: any, c: ColumnParameter) {
        let mapping = this.rowMappings[c.accessor || ''] || '';
        if (mapping[c.aggregation || '']) {
            mapping = mapping[c.aggregation || '']
        }
        if (mapping && !mapping.split && (Object.keys(mapping).length === 1)) {
            mapping = mapping[Object.keys(mapping)[0]]
        }
        if (mapping && mapping.split) {
            let obj = row;
            mapping.split('.').forEach((step: string) => {
                if (isArrayLike(obj)) {
                    obj = obj.map((o: any) => {
                        if (o) {
                            return o[step];
                        } else {
                            return o;
                        }
                    });
                } else {
                    if (obj) {
                        obj = obj[step]
                    }
                }
            });
            return obj;
        } else {
            return null;
        }
    }

    @action.bound
    drillPage(key: string, drillValues: any) {
        if (this.drillPages[key]) {
            runInAction(() => {
                this.drillPages[key].component = this.component.drillOne(drillValues);
            });
        } else {
            this.drillPages[key] = new PaginatedReportComponent(this.component.drillOne(drillValues));
        }

        const paginatedChild = this.drillPages[key];
        paginatedChild.cloneDates(this);
        runInAction(() => {
            paginatedChild.resizings = this.resizings;
        });
        if (paginatedChild.viewModel.isDirty) {
            paginatedChild.commit();
        }
        return paginatedChild;
    }

    @action.bound
    flagRanSizing() {
        this.uiRanSizing = true;
    }

    @action.bound
    flagRenderedTotals() {
        this.uiRenderedTotals = true;
    }

    @action.bound
    setFontSize(newSize: number) {
        this.fontSize = newSize;
    }

    columnID = computedFn((c: ColumnParameter) => {
        return `${c.accessor}.${c.aggregation}`
    })

    @computed get fontString() {
        return `${this.fontSize}px Calibri, sans-serif`
    }

    @action.bound
    setResizings(newResizings: Resize[]) {
        this.resizings = (this.component.viewModel.parameters.columns || []).map((c: ColumnParameter) => {
            const id = this.columnID(c);
            return newResizings.find(r => r.id === id) || {id: id, value: 150}
        });
        this.flagRanSizing();
    }

    @action.bound
    setExpansions(newExpansions: { [i: number]: {} }) {
        this.expansions = newExpansions
    }

    @action.bound
    expandAll() {
        this.collapseAll();
        for (let i = 0; i < this.pageSize; i++) {
            this.expansions[i] = {};
        }
    }

    @action.bound
    collapseAll() {
        this.expansions = {};
    }

    @computed get fullyExpanded() {
        for (let i = 0; i < this.pageSize; i++) {
            if (!this.expansions[i]) {
                return false;
            }
        }
        return true;
    }

    @computed
    get reactTableSorts() {
        // Why would you do this in such a strange way?
        // Because the way that react table checks if the old sorts !== the new sorts on a render means they need to be exactly the same object or a new data fetch happens
        // And apparently the way mobx was computing this meant a data fetch on every render for a subtable
        // And the way the width was being calculated that meant data fetches on the window scroll event for drill tables
        const tmp = ((this.component.viewModel.parameters.columns || []).slice() as ColumnParameter[])
            .filter(c => c.sort_priority || (c.sort_priority === 0))
            .sort((a, b) => a.sort_priority! - b.sort_priority!).map(c => ({
                id: this.columnID(c),
                desc: !!c.sort_desc
            }));
        if ((this.internalRTSorts.length !== tmp.length) || this.internalRTSorts.some((s, i) => {
            return !tmp[i] || (tmp[i].id !== s.id) || (tmp[i].desc !== s.desc)
        })) {
            this.internalRTSorts = tmp;
        }
        return this.internalRTSorts;
    }

    @action.bound
    setReactTableSorts(newSorts: { id: string, desc: boolean }[]) {
        this.component.nestedAssignment('parameters.columns', (this.component.viewModel.parameters.columns || []).map((c: ColumnParameter) => {
            let matchingSort = newSorts.findIndex(s => s.id === this.columnID(c));
            return {
                ...c,
                sort_priority: (matchingSort === -1) ? undefined : matchingSort + 1,
                sort_desc: (matchingSort === -1) ? undefined : newSorts[matchingSort].desc
            }
        }));
    }

    runFilterValue = computedFn(function getVal(this: PaginatedReportComponent, columnName: string) {
        return this.viewModel.runFilters[columnName];
    });

    @action.bound
    handleRunFilterChange(name: string | undefined, val: string | string[]) {
        if(!isArrayLike(val)){
            const existing = this.runFilterValue(name!) || []
            val = [...existing, val];
        }
        this.viewModel.runFilters = {...this.viewModel.runFilters, [name!]: val}
    }
}
