import {attribute, model, Model, Transport} from '@rx-suite/data';
import {i18n} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import * as yup from "yup";
import {observable} from "mobx";
import FileSaver from 'file-saver'
import {PaginatedReportComponent} from "./PaginatedReportComponent";

class ReportComponentResultTransport<T extends Model> extends Transport<T> {
    async run(parameters: any) {
        const resp = await this.fetch(`/run`, {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({
                ...parameters,
                appId: localStorage.getItem('appId')
            })
        });
        return resp.json();
    }

    async rerun(id: string) {
        const resp = await this.fetch(`/${id}/rerun`, {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            }
        });
        return resp.json();
    }

    async download(link: string) {
        const resp = await ReportComponentResultTransport.fetch(link, {method: 'GET'});
        if (resp.ok) {
            return resp.blob();
        }
    }
}

@model({
    transportLayer: new ReportComponentResultTransport("/api/report_component_results", "report_component_result"),
    permissionTarget: "ReportComponentResult",
    routingInfo: {
        read: "/report_component_results/:id/edit",
    }
})
export class ReportComponentResultModel extends Model {
    @attribute({
        label: i18n._(t`Name`),
        validation: yup.string().required(i18n._(t`You must give your result a name`))
    })
    name: string = '';

    @attribute({
        label: i18n._(t`Format`),
        validation: yup.string().required(i18n._(t`You must select a result format`))
    })
    data_format: string = '';

    @attribute({
        label: i18n._(t`Description`),
        validation: yup.string()
    })
    desc: string = '';

    @attribute({
        label: i18n._(t`Generating?`),
        validation: yup.boolean()
    })
    generating: boolean = true;

    @attribute({
        label: i18n._(t`Completion`),
        validation: yup.number().nullable()
    })
    generation_percent?: number | null = null;

    @attribute({
        label: i18n._(t`Error`),
        validation: yup.string().nullable()
    })
    generation_error?: string | null = null;

    @attribute({
        label: i18n._(t`Parameters`),
        validation: yup.string(),
        autoTable: false
    })
    parameters: any = observable({});

    @attribute({
        label: i18n._(t`Target`),
        connects: 'ReportTargetModel',
        validation: yup.number().required(i18n._(t`You must set a report target`)),
        autoTable: false
    })
    report_target_id: number | undefined = undefined;

    @attribute({label: i18n._(t`Created At`), autoForm: true})
    created_at: Date = new Date();

    @attribute({
        label: i18n._(t`Download`),
        validation: yup.string().nullable()
    })
    download_link?: string | null = null;

    download = () => {
        if (this.download_link) {
            FileSaver.saveAs(this.download_link)
        }
    };

    run = () => {
        (this.store.transportLayer as ReportComponentResultTransport<ReportComponentResultModel>).rerun(this.id as string);
    }
}

export const ReportComponentResultStore = ReportComponentResultModel.store;

export async function makeExport(paginatedComponent: PaginatedReportComponent, data_format: string, overrideParameters?: any) {
    const transport = ReportComponentResultStore.transportLayer as ReportComponentResultTransport<ReportComponentResultModel>;
    return transport.run({
        ...paginatedComponent.component.viewModel.asJson,
        parameters: {
            ...paginatedComponent.component.viewModel.parameters,
            filter: paginatedComponent.component.filterWithRunParams(paginatedComponent.viewModel.dateColumn, paginatedComponent.viewModel.startDate, paginatedComponent.viewModel.endDate, paginatedComponent.viewModel.runFilters),
            ...(overrideParameters || {})
        },
        data_format: data_format
    })
}
