import {attribute, Model, model, Transport, Types} from '@rx-suite/data';
import * as yup from 'yup';
import {i18n} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import {action, computed, observable, runInAction, when} from "mobx";
import {computedFn} from "mobx-utils";
import {ReportTargetModel, ReportTargetStore} from "./ReportTargetStore";
import {ColumnParameter} from "./Types";
import moment, {Moment} from "moment";

@model({
    transportLayer: new Transport("/api/report_components", "report_component"),
    permissionTarget: "ReportComponent",
    routingInfo: {
        list: '/report_components',
        create: '/report_components/create',
        read: '/report_components/:id/read?tab=runner',
        update: '/report_components/:id/edit?tab=builder'
    }
})
export class ReportComponentModel extends Model {
    types: Types = {
        parameters: 'json',
    };

    @attribute({
        label: i18n._(t`Name`),
        description: i18n._(t`This is how you will identify the component in lists (for example, you can search for this name).  It will also be shown as a title on some component types.`),
        validation: yup.string().required(i18n._(t`You must give your component a name`))
    })
    name: string = '';

    @attribute({
        label: i18n._(t`Description`),
        description: i18n._(t`Add any notes here.  The information will not be included in the report display - it is for reference purposes only.`),
        validation: yup.string()
    })
    desc: string = '';

    @attribute({
        label: i18n._(t`Canned Id`),
        validation: yup.string().nullable(),
        autoTable: false,
        autoForm: false
    })
    canned_report_id?: string | null = null;

    @attribute({
        label: i18n._(t`Parameters`),
        validation: yup.string(),
        autoTable: false
    })
    parameters: any = observable({
        columns: [],
    });

    @computed
    get printParameters() {
        return {
            ...this.viewModel.parameters,
            columns: this.printColumns
        };
    }

    @computed
    get printColumns() {
        if (this.hasGroupingColumn) {
            return (this.viewModel.parameters.columns as ColumnParameter[]).filter(c => !!c.aggregation);
        } else {
            return this.viewModel.parameters.columns;
        }
    }

    @attribute({
        label: i18n._(t`Target`),
        connects: 'ReportTargetModel',
        description: i18n._(t`The selected target determines what kinds of data can be fetched and what options you have to display it.  Available targets depend on your installed Rx-Suite modules.`),
        validation: yup.number().required(i18n._(t`You must set a report target`)),
        autoTable: false
    })
    report_target_id: number | undefined = undefined;
    getReturnDef = computedFn(function getDef(this: ReportComponentModel, accessor: string | undefined): any {
        if (!accessor) {
            return null
        }
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.returns || !target.returns.properties) {
            return false;
        }
        let columnInfo = target.returns.properties;
        const path = accessor.split('.');
        path.forEach((step: string, i: number) => {
            if (columnInfo) {
                columnInfo = columnInfo[step];
            }
            if (columnInfo && columnInfo.subdefinition && i !== (path.length - 1)) {
                columnInfo = columnInfo.subdefinition;
            }
        });
        return columnInfo;
    });
    internalGroupingNameForColumn = computedFn(function getDef(this: ReportComponentModel, accessor: string, depth: number): any {
        const def = this.getReturnDef(accessor);
        if (def && def.groupname) {
            return def.groupname;
        }
        const parentPath = accessor.split('.');
        parentPath.pop();
        let result;
        if (parentPath.length > 0) {
            result = this.internalGroupingNameForColumn(parentPath.join('.'), depth + 1);
        }
        return result || (depth === 1 ? def.name : '');
    });
    groupingNameForColumn = computedFn(function getDef(this: ReportComponentModel, accessor: string): any {
        return this.internalGroupingNameForColumn(accessor, 0);
    });
    isGroupingOperation = computedFn(function getDef(this: ReportComponentModel, aggregation?: string): boolean {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.ui_hints) {
            return false;
        }
        return !!Object.entries(target.ui_hints.grouping_operations || {}).find(([id, g]) => (
            id === aggregation
        ));
    });
    aggregationOperationsForColumn = computedFn(function getDef(this: ReportComponentModel, accessor: string | undefined, skipGrouped: boolean): any {
        const target = this.viewModel.report_target as ReportTargetModel;
        let columnInfo = this.getReturnDef(accessor);
        if (!columnInfo || !target || !target.ui_hints) {
            return [];
        }
        const results: { id: string, name: string }[] = [];
        Object.entries(target.ui_hints.aggregation_operations || {}).filter(([id, a]) => {
            if (a.applicable_to && !a.applicable_to.find(type => type === columnInfo.type)) {
                return false;
            } else if (a.only_for_groups && skipGrouped) {
                return false;
            }
            return true;
        }).forEach(([id, a]) => {
            results.push({id: id, name: a.name})
        });
        return results;
    });
    groupingOperationsForColumn = computedFn(function getDef(this: ReportComponentModel, accessor: string): any {
        const target = this.viewModel.report_target as ReportTargetModel;
        let columnInfo = this.getReturnDef(accessor);
        if (!columnInfo || !target || !target.ui_hints || this.columnIsOneToMany(accessor)) {
            return [];
        }
        const results: { id: string, name: string }[] = [];
        Object.entries(target.ui_hints.grouping_operations || {}).forEach(([id, g]) => {
            if (!g.applicable_to || g.applicable_to.find(type => type === columnInfo!.type)) {
                results.push({id: id, name: g.name})
            }
        });
        return results;
    });
    columnIsOneToMany = computedFn(function getDef(this: ReportComponentModel, accessor: string): any {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.returns || !target.returns.properties) {
            return false;
        }
        let columnInfo = target.returns.properties;
        let foundMultiple = false;
        const path = accessor.split('.');
        path.forEach((step: string, i: number) => {
            if (columnInfo) {
                if(columnInfo.multiple){
                    foundMultiple = true
                }
                columnInfo = columnInfo[step];
            }
            if (columnInfo && columnInfo.subdefinition && i !== (path.length - 1)) {
                if(columnInfo.multiple){
                    foundMultiple = true
                }
                columnInfo = columnInfo.subdefinition;
            }
        });
        return foundMultiple;
    });
    aggregationDetails = computedFn(function getDef(this: ReportComponentModel, aggregation?: string): any {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.ui_hints || !aggregation) {
            return [];
        }
        return (target.ui_hints.aggregation_operations || {})[aggregation] || (target.ui_hints.grouping_operations || {})[aggregation]
    });
    withExtraFilters = computedFn(function getFilter(this: ReportComponentModel, criteria: any[]): any {
        const existingParameters = this.viewModel.parameters || {};
        const existingFilter = existingParameters.filter;
        const filters = [existingFilter, ...criteria].filter(f => f);
        return filters.length ? {"and": filters} : {}
    });
    filterWithRunParams = computedFn(function getFilter(this: ReportComponentModel, dateColumn: string | null, startDate: Date | string | Moment | null, endDate: Date | string | Moment | null, runParams: { [key: string]: string[] } | null): any {
        const afterFilter = (dateColumn && startDate) ? {">=": [{"var": dateColumn}, moment(startDate).toISOString()]} : null;
        const beforeFilter = (dateColumn && endDate) ? {"<=": [{"var": dateColumn}, moment(endDate).toISOString()]} : null;
        const extraFilters: any[] = [];
        if (runParams) {
            Object.entries(runParams).forEach(([key, value]) => {
                if (value && value.length) {
                    extraFilters.push({
                        "==": [
                            {"var": key},
                            value
                        ]
                    });
                }
            });
        }
        return this.withExtraFilters([afterFilter, beforeFilter, ...extraFilters])
    });

    @computed get supportsLiveView() {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.ui_hints) {
            return false;
        }
        return !!target.ui_hints.live_pagination
    }

    @computed get supportsCSVDump() {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.ui_hints) {
            return false;
        }
        return (target.ui_hints.type === 'automodel')
    }

    @computed
    get groupingColumns() {
        return (this.viewModel.parameters.columns || []).filter((c: ColumnParameter) => this.isGroupingOperation(c.aggregation)) as ColumnParameter[]
    }

    @computed
    get primeGroupingColumn() {
        return this.groupingColumns[0];
    }

    // because of how we set nested groups up as subcomponents, we need to feed a query with subgroups stripped
    @computed
    get columnsWithSingleGroup() {
        let foundGroup = false;
        return (this.viewModel.parameters.columns || []).map((c: ColumnParameter) => {
            const proc = JSON.parse(JSON.stringify(c));
            if (this.isGroupingOperation(proc.aggregation)) {
                if (foundGroup) {
                    proc.aggregation = '';
                }
                foundGroup = true;
            }
            return proc;
        })
    }

    @computed
    get hasGroupingColumn() {
        return this.groupingColumns.length > 0
    }

    @computed get supportedDateRangeColumns() {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target) {
            return {};
        }
        const dates = Object.entries(target.returns?.properties || {}).filter(([_key, c]: [string, any]) => c.type === 'date' || c.type === 'datetime' || c.type === 'time').reduce((bucket: any, [key, c]: [string, any]) => {
            bucket[key] = c;
            return bucket;
        }, {});
        (target.ui_hints?.offer_dates_from || []).forEach(accessor => {
            let def = target.returns?.properties || {};
            accessor.split('.').forEach(step => {
                if (def) {
                    def = def[step]
                }
                if (def && def.subdefinition) {
                    def = def.subdefinition
                }
            });
            if (def) {
                Object.entries(def).filter(([_key, c]: [string, any]) => c.type === 'date' || c.type === 'datetime' || c.type === 'time').forEach(([key, c]: [string, any]) => {
                    dates[`${accessor}.${key}`] = c;
                });
            }
        });
        return dates;
    }

    @computed get supportedRunFilters() {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target || !target.ui_hints || !target.ui_hints.runner_params) {
            return {};
        }
        return target.ui_hints.runner_params.reduce((bucket: any, accessor: string) => {
            const returnDef = this.getReturnDef(accessor);
            if (returnDef) {
                bucket[accessor] = returnDef;
            }
            return bucket;
        }, {});
    }

    @computed
    get exportName() {
        return this.name || this.getId() as string;
    }

    @computed
    get isUserDeletable() {
        return (!this.canned_report_id) && this.designPermitted;
    }

    @computed
    get isUserUpdatable() {
        return this.designPermitted;
    }

    @computed
    get isUserReadable() {
        return this.runPermitted;
    }

    @computed get referencedTargets(): ReportTargetModel[] {
        const additionalTargets = new Set();
        (this.viewModel.parameters.columns || []).forEach((c: ColumnParameter) => {
            if (c && c.accessor) {
                const path = c.accessor.split('.');
                if (path.length > 1) {
                    const lastTargetID = (this.getReturnDef(path.slice(0, path.length - 1).join('.')) || {}).subreportid; //recall: the final item (length - 1) is a column; the one before it is the resolved table
                    if (lastTargetID) {
                        additionalTargets.add(ReportTargetStore.models.find(t => t.report_id === lastTargetID))
                    }
                }
            }
        });
        return [this.report_target, ...additionalTargets].filter(t => t) as ReportTargetModel[];
    }

    @computed get runPermitted() {
        return this.referencedTargets.every((t: ReportTargetModel) => t.runPermitted)
    }

    @computed get designPermitted() {
        return !this.canned_report_id && this.referencedTargets.every((t: ReportTargetModel) => t.designPermitted)
    }

    @computed get accessPermitted() {
        return this.referencedTargets.every((t: ReportTargetModel) => t.accessPermitted)
    }

    @action.bound
    validateGroupOnlyAggregations() {
        if (!this.hasGroupingColumn) {
            (this.viewModel.parameters.columns || []).forEach((c: ColumnParameter) => {
                if ((this.aggregationDetails(c.aggregation) || {}).only_for_groups) {
                    c.aggregation = '';
                }
            });
        }
    }

    drillOne(groupValues: any): ReportComponentModel {
        if (this.hasGroupingColumn) {
            const child = ReportComponentStore.cloneModel(this) as ReportComponentModel;
            child.nestedAssignment('parameters', JSON.parse(JSON.stringify(this.viewModel.parameters)));
            const filters = this.groupingColumns.reduce((f: any, c: ColumnParameter) => {
                if (c.accessor && groupValues.hasOwnProperty((c.accessor))) {
                    switch (c.aggregation) {
                        case 'group_by_day':
                            f.push({">=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).startOf('day')]});
                            f.push({"<=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).endOf('day')]});
                            break;
                        case 'group_by_week':
                            f.push({">=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).startOf('week')]});
                            f.push({"<=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).endOf('week')]});
                            break;
                        case 'group_by_month':
                            f.push({">=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).startOf('month')]});
                            f.push({"<=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).endOf('month')]});
                            break;
                        case 'group_by_quarter':
                            f.push({">=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).startOf('quarter')]});
                            f.push({"<=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).endOf('quarter')]});
                            break;
                        case 'group_by_year':
                            f.push({">=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).startOf('year')]});
                            f.push({"<=": [{"var": c.accessor}, moment.parseZone(groupValues[c.accessor]).endOf('year')]});
                            break;
                        case 'group_with_column':
                            const values = groupValues[c.accessor].split("&|>");
                            [c.accessor, ...(c.aggregation_options?.join_columns || [])].forEach((accessor, i) => {
                                const value = values[i];
                                if (value == '__') {
                                    f.push({"==": [{"var": accessor}, null]});
                                } else {
                                    f.push({"==": [{"var": accessor}, value]});
                                }
                            });
                            break;
                        case 'group_by_same_value':
                        default:
                            f.push({"==": [{"var": c.accessor}, groupValues[c.accessor]]});
                            break;
                    }
                }
                return f;
            }, []);
            child.nestedAssignment(`parameters.filter`, this.withExtraFilters(filters));
            child.removeColumn(child.groupingColumns[0]);
            return child;
        } else {
            return ReportComponentStore.cloneModel(this) as ReportComponentModel;
        }
    }

    @action
    setColumnProp(item: ColumnParameter, prop: string, newVal: any) {
        const index = this.viewModel.parameters.columns.findIndex((c: ColumnParameter) => c === item);
        this.nestedAssignment(`parameters.columns[${index}].${prop}`, newVal);
        this.validateGroupOnlyAggregations();
    }

    getData = async (overrideParameters?: any) => {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target) {
            return null;
        }
        return target.getData(Object.assign({}, this.viewModel.parameters, {grand_totals: false}, overrideParameters || {}));
    };

    getTotals = async (overrideParameters?: any) => {
        const target = this.viewModel.report_target as ReportTargetModel;
        if (!target) {
            return null;
        }
        const resolvedParameters = Object.assign({}, this.viewModel.parameters, overrideParameters || {}, {grand_totals: true});
        if (!resolvedParameters.columns.some((c: ColumnParameter) => c.aggregation)) {
            return null;
        }
        return target.getData(resolvedParameters);
    };

    groupedColumnsAtDepth(level: number) {
        const selectedGroups = this.groupingColumns.slice(0, level + 1).map(c => c.accessor);
        let foundGroup = false;
        return (this.viewModel.parameters.columns as ColumnParameter[] || []).map(c => {
            let aggregation = c.aggregation;
            if (selectedGroups.includes(c.accessor)) {
                if (c.accessor === selectedGroups[selectedGroups.length - 1]) {
                    foundGroup = true;
                }
            } else if (foundGroup && this.isGroupingOperation(c.aggregation)) {
                aggregation = '';
            }
            return {
                ...c,
                aggregation: aggregation
            }
        });
    }

    @computed
    get columnsUngrouped() {
        return (this.viewModel.parameters.columns as ColumnParameter[] || []).map(c => {
            return {
                ...c,
                aggregation: null
            }
        });
    }

    @action.bound
    removeColumn(column: any) {
        this.nestedAssignment('parameters.columns', (this.viewModel.parameters.columns || []).filter((c: any) => c !== column));
    }

    asExport(): object {
        const standardDump = super.asExport() as any;
        const target = this.report_target as ReportTargetModel;
        const result = {
            report_target: {
                module_id: target.module_id,
                report_id: target.report_id,
            }
        } as any;
        Object.keys(standardDump).forEach(k => {
            if (k !== 'report_target_id' && k !== 'canned_report_id') {
                result[k] = standardDump[k]
            }
        });
        return result;
    }

    /**
     * Inverse of asExport
     * @returns {object} An export-friendly object
     */
    @action
    fromExport(json: Model) {
        this.updateFromJson(json, true);
        if (json.report_target) {
            when(() => ReportTargetStore.models.length > 0, () => {
                runInAction(() => {
                    const target_data = json.report_target as { module_id?: string, report_id?: string };
                    const target = ReportTargetStore.models.find(t => t.module_id === target_data.module_id && t.report_id === target_data.report_id);
                    if (target) {
                        this.report_target_id = target.id as number;
                    }
                })
            });
        }
    }
}

export const ReportComponentStore = ReportComponentModel.store;
