import {attribute, DataStore, Model, model, Transport, Types} from '@rx-suite/data';
import {Parameters, UIHints} from './Types';
import * as yup from 'yup';
import {i18n} from "@rx-suite/ui";
import {t} from "@lingui/macro";
import {computed} from "mobx";
import {currentUserCanModule, featureEnabled} from "@rx-suite/stores";

class ReportTargetTransport<T extends Model> extends Transport<T> {
    async getReportData(id: string, parameters: any) {
        const resp = await this.fetch(`/${id}/run`, {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({
                parameters,
                appId: localStorage.getItem('appId')
            })
        });
        return resp.json();
    }
    async runFilterOptions(id: string, column: string) {
        const resp = await this.fetch(`/${id}/${column.replace(/\./g, '-')}/known_filter_vals`, {
            method: 'GET',
            headers: {
                'Content-Type': 'application/json'
            }
        });
        return resp.json();
    }
}

@model({
    transportLayer: new ReportTargetTransport("/api/report_targets", "report_target"),
    eager: true
})
export class ReportTargetModel extends Model {
    types: Types = {
        ui_hints: 'json',
        parameters: 'json',
        returns: 'json',
    };

    @attribute({
        validation: yup.string()
    })
    module_id: string = '';

    @attribute({
        label: i18n._(t`Module`),
        validation: yup.string()
    })
    module_name: string = '';

    @attribute({
        validation: yup.string()
    })
    report_id: string = '';

    @attribute({
        validation: yup.string()
    })
    permission_id: string = '';

    @attribute({
        label: i18n._(t`Report Name`),
        validation: yup.string()
    })
    name: string = '';

    @attribute({
        label: i18n._(t`Report Description`),
        validation: yup.string()
    })
    desc: string = '';

    @attribute({
        validation: yup.mixed()
    })
    ui_hints: UIHints = {};

    @attribute({
        validation: yup.mixed()
    })
    parameters: Parameters = {};

    @attribute({
        validation: yup.mixed()
    })
    returns: any = null;

    @computed get runPermitted() {
        return currentUserCanModule(this.permission_id || this.report_id, 'run', this.module_id);
    }

    @computed get designPermitted() {
        return featureEnabled('designer') && currentUserCanModule(this.permission_id || this.report_id, 'design', this.module_id);
    }

    @computed get accessPermitted() {
        return this.runPermitted || this.designPermitted;
    }

    getData = async (parameters: any) => {
        const constructor = this.constructor as unknown as { store: DataStore<ReportTargetModel> };
        const store = constructor.store.transportLayer as ReportTargetTransport<ReportTargetModel>;
        return store.getReportData(this.getId(), parameters);
    };

    runFilterOptions = async (column: string) => {
        const constructor = this.constructor as unknown as { store: DataStore<ReportTargetModel> };
        const store = constructor.store.transportLayer as ReportTargetTransport<ReportTargetModel>;
        const resp = await store.runFilterOptions(this.getId(), column);
        return resp.values.filter((v:any) => v !== null);
    };
}

export const ReportTargetStore = ReportTargetModel.store;
ReportTargetStore.setLimit(10000);
