# == Schema Information
#
# Table name: automatic_reports
#
#  id                  :bigint           not null, primary key
#  attach_csv          :boolean
#  attach_link         :boolean
#  attach_pdf          :boolean
#  computed_next_run   :datetime
#  date_column         :string
#  last_run_error      :string
#  last_run_time       :datetime
#  mail_to_emails      :string
#  mail_to_users       :string
#  name                :string
#  period_unit         :string
#  period_value        :integer
#  run_frequency_unit  :string
#  run_frequency_value :integer
#  run_params          :string
#  run_token           :string
#  start_date          :datetime
#  created_at          :datetime         not null
#  updated_at          :datetime         not null
#  pdf_template_id     :string
#  report_component_id :bigint           not null
#
# Indexes
#
#  index_automatic_reports_on_report_component_id  (report_component_id)
#
# Foreign Keys
#
#  fk_rails_...  (report_component_id => report_components.id)
#
class AutomaticReport < ApplicationRecord
  belongs_to :report_component
  has_one :report_target, through: :report_component

  after_commit :log_schedule

  def log_schedule
    if previous_changes.include?(:run_frequency_value) ||
        previous_changes.include?(:run_frequency_unit) ||
        previous_changes.include?(:period_value) ||
        previous_changes.include?(:period_unit) ||
        previous_changes.include?(:start_date)
      Rails.logger.info do
        {
          msg: 'Automatic reporting schedule changed',
            report: id,
            user: RxModule::Authorization.current_user&.[](:name),
            changes: {
              run_frequency_value: previous_changes[:run_frequency_value],
                run_frequency_unit: previous_changes[:run_frequency_unit],
                period_value: previous_changes[:period_value],
                period_unit: previous_changes[:period_unit],
                start_date: previous_changes[:start_date],
            }
        }
      end
      set_next_run!
    end
  end

  def self.display_name
    'Automatic Report'
  end

  def unit_for_advance
    case run_frequency_unit.to_sym
    when :day, :days
      :days
    when :week, :weeks
      :weeks
    when :month, :months
      :months
    when :year, :years
      :years
    else
      :days
    end
  end

  def unit_for_regress
    case period_unit.to_sym
    when :day, :days
      :days
    when :week, :weeks
      :weeks
    when :month, :months
      :months
    when :year, :years
      :years
    else
      :days
    end
  end

  def rufus_freq
    "#{run_frequency_value}#{run_frequency_unit}"
  end

  def frequency(val = nil)
    begin
      (val || run_frequency_value).send(run_frequency_unit)
    rescue StandardError => err
      nil
    end
  end

  def compute_next_run
    Rails.logger.debug { {msg: 'Computing next run',
                         run_frequency_value: run_frequency_value,
                         run_frequency_unit: run_frequency_unit,
                         period_value: period_value,
                         period_unit: period_unit,
                         start_date: start_date,
                         report: id} }
    if start_date >= RxModule::Time.now
      Rails.logger.debug { {msg: 'Computing next run: start date is in the future', start_date: start_date, report: id} }
      return start_date
    else
      timestamp_diff = (RxModule::Time.now.to_i - start_date.to_i).abs
      scaled_diff = (timestamp_diff / frequency.to_i)
      intermediate_date = start_date.advance(unit_for_advance => scaled_diff)
      if (intermediate_date <= RxModule::Time.now) && last_run_time && (intermediate_date <= last_run_time)
        new_date = intermediate_date.advance(unit_for_advance => run_frequency_value)
        Rails.logger.debug { {msg: 'Computing next run: rolling forward one rounding unit', predate: intermediate_date, postdate: new_date, report: id} }
        intermediate_date = new_date
      end
      if last_run_error && last_run_error != '' && last_run_time && intermediate_date >= last_run_time.advance(minutes: 10)
        new_date = last_run_time.advance(minutes: 10)
        Rails.logger.debug { {msg: 'Computing next run: last run was error & computed date too far in future.  Rolling forward from last time', last_run_time: last_run_time, predate: intermediate_date, postdate: new_date, report: id} }
        intermediate_date = new_date
      end
      if intermediate_date <= Time.now
        new_date = Time.now
        Rails.logger.debug { {msg: 'Computing next run: computed date is in the past, rolling forward', predate: intermediate_date, postdate: new_date, report: id} }
        intermediate_date = new_date
      end
      intermediate_date
    end
  end

  def set_next_run!
    new_run = compute_next_run
    update(computed_next_run: new_run) if computed_next_run != new_run
  end

  def generate(end_date)
    begin
      RxModule::Authorization.set_request_token(run_token)
      subject = "#{RxModule::Localization.translate("Automatic Report Completed")}: #{report_component.name}"
      end_date ||= Time.now
      start_date = end_date.advance(unit_for_regress => -1 * (period_value || 0))
      parsed_run_params = {}
      begin
        parsed_run_params = JSON.parse(run_params)
        parsed_run_params = {} unless parsed_run_params.is_a?(Hash)
      rescue StandardError
        # ignored
      end
      url = report_component.run_url(date_column: date_column, end_date: end_date, start_date: start_date, run_params: parsed_run_params)

      Rails.logger.info { {msg: 'Automatic report running', start: start_date, end: end_date, url: url, report: as_json, as: RxModule::Authorization.current_user&.[](:name)} }
      update(last_run_time: Time.now)

      body_text = {
        preamble: [
            RxModule::Localization.translate('An automatically generated report is ready to be viewed.' + '<br/>'),
            report_component.name ? (RxModule::Localization.translate('Report Name: ') + report_component.name) : nil,
            name ? (RxModule::Localization.translate('Automatic Report Name: ') + name) : nil,
            date_column ? ('<br/>' + RxModule::Localization.translate('Date Type: ') + report_target.display_name(date_column)) : nil,
            start_date ? (RxModule::Localization.translate('Start Date: ') + start_date.strftime('%FT%T')) : nil,
            end_date ? (RxModule::Localization.translate('End Date: ') + end_date.strftime('%FT%T')) : nil,
            '<br/>' + RxModule::Localization.translate('Run Date: ') + Time.now.strftime('%FT%T'),
        ].compact.join('<br/>'),
          url: nil,
          csv: nil,
          pdf: nil
      }

      if attach_link
        body_text[:url] = "#{RxModule::Localization.translate("Link")}: <a href=\"#{url}\">#{url}</a>"
      end

      attachment_data = []
      if attach_csv
        body_text[:csv] = RxModule::Localization.translate('A CSV file has been attached')
        result = report_component.make_csv(date_column, start_date, end_date, parsed_run_params)
        attachment_data.push({path: result.data_file_disk_path, format: 'text/csv', name: result.filename})
      end

      pdf_file = nil
      if attach_pdf
        body_text[:pdf] = RxModule::Localization.translate('A PDF file has been attached')
        result = report_component.get_pdf_result(pdf_template_id, date_column, start_date, end_date, true, parsed_run_params)
        response = Faraday.get(result[:download_link]) do |req|
          req.headers[:Authorization] = "TOKEN #{RxModule::Authorization.current_user_token}"
        end
        pdf_file = Tempfile.create(mode: File::Constants::BINARY)
        pdf_file.write(response.body)
        pdf_file.flush
        attachment_data.push({path: pdf_file.path, format: 'application/pdf', name: "#{report_component.name || 'result'}.pdf"})
      end

      user_data = RxModule::Communicator::Hub.picker_users
      user_emails = begin
                      JSON.parse(mail_to_users).map do |u_id|
                        user_data.find do |u_d|
                          u_d[:id] == u_id
                        end&.[](:email)
                      end
                    rescue StandardError => err
                      []
                    end.compact
      if user_emails.length > 0
        body = [body_text[:preamble], body_text[:url], body_text[:csv], body_text[:pdf]].compact.join('<br/><br/>')
        RxModule::Communicator::Emails.send_email(user_emails, subject, body, attachment_data.map { |a| RxModule::Communicator::Emails.attachable_file(a[:path], a[:format], a[:name]) })
      end
      exterior_emails = begin
                          JSON.parse(mail_to_emails)
                        rescue StandardError => err
                          []
                        end.compact
      if exterior_emails.length > 0
        body = [body_text[:preamble], body_text[:csv], body_text[:pdf]].compact.join('<br/><br/>')
        RxModule::Communicator::Emails.send_email(exterior_emails, subject, body, attachment_data.map { |a| RxModule::Communicator::Emails.attachable_file(a[:path], a[:format], a[:name]) })
      end
      begin
        pdf_file&.close
        pdf_file&.unlink
      rescue StandardError => err
        Rails.logger.debug { {msg: 'Failed to explicitly delete tempfile; will be garbage collected later', report: as_json} }
      end
      update(last_run_error: nil)
      set_next_run!
    rescue StandardError => err
      update(last_run_error: err.message)
      Rails.logger.error { {msg: 'Automatic report failed to run', report: as_json} }
    end
  end
end
