# == Schema Information
#
# Table name: report_components
#
#  id               :bigint           not null, primary key
#  desc             :string
#  name             :string
#  parameters       :string
#  created_at       :datetime         not null
#  updated_at       :datetime         not null
#  canned_report_id :string
#  report_target_id :bigint
#
# Indexes
#
#  index_report_components_on_report_target_id  (report_target_id)
#
# Foreign Keys
#
#  fk_rails_...  (report_target_id => report_targets.id) ON DELETE => cascade ON UPDATE => cascade
#
class ReportComponent < ApplicationRecord
  belongs_to :report_target

  class << self
    def display_name
      "Report Component"
    end
  end

  def params_with_date_filter(date_column = nil, start_date = nil, end_date = nil)
    params = begin
               JSON.parse(parameters)
             rescue StandardError => err
               nil
             end || {}
    if date_column
      params["filter"] = {
          "and" => [
              start_date ? {">=" => [{"var" => date_column}, start_date.strftime("%FT%T")]} : nil,
              end_date ? {"<=" => [{"var" => date_column}, end_date.strftime("%FT%T")]} : nil,
              params["filter"]
          ].compact
      }
    end
    params
  end

  def params_with_run_params(existing_params, run_params)
    return existing_params unless run_params
    new_params = existing_params.dup
    new_params["filter"] = {
        "and" => [
            *run_params.map do |k, v|
              {"==" => [{"var" => k}, v]}
            end,
            existing_params["filter"]
        ].compact
    }
    new_params
  end

  def make_csv(date_column = nil, start_date = nil, end_date = nil, run_params = nil)
    filter_params = JSON.generate(params_with_run_params(params_with_date_filter(date_column, start_date, end_date), run_params))
    Rails.logger.info { {msg: "New CSV Result", date_column: date_column, start_date: start_date, end_date: end_date, filter_params: filter_params, report: as_json} }
    result = ReportComponentResult.create(name: name, parameters: filter_params, desc: desc, data_format: 'csv', report_target_id: report_target.id)
    result.save
    result.generate_file
    result
  end

  def run_url(opts = {})
    date_format_str = "%FT%T"
    relative_url = "/report_components/#{id}/read?tab=runner"
    relative_url = "#{relative_url}&datetype=#{Rack::Utils.escape(opts[:date_column])}" if opts[:date_column]
    relative_url = "#{relative_url}&startdate=#{Rack::Utils.escape(opts[:start_date].strftime(date_format_str))}" if opts[:start_date]
    relative_url = "#{relative_url}&enddate=#{Rack::Utils.escape(opts[:end_date].strftime(date_format_str))}" if opts[:end_date]
    relative_url = "#{relative_url}&pageNumber=#{Rack::Utils.escape(opts[:page_number])}" if opts[:page_number]
    relative_url = "#{relative_url}&pageSize=#{Rack::Utils.escape(opts[:page_size])}" if opts[:page_size]
    relative_url = "#{relative_url}&printView=#{Rack::Utils.escape(opts[:print_view].to_s)}" if opts[:print_view]
    relative_url = "#{relative_url}&isLastPage=#{Rack::Utils.escape(opts[:is_last_page].to_s)}" if opts[:is_last_page]
    opts[:run_params]&.each do |k, v|
      relative_url = "#{relative_url}&#{k}=#{(v || []).map { |sub_v| Rack::Utils.escape(sub_v.to_s) }.join('|')}"
    end

    RxModule::Communicator.resolve_endpoint(RxModule.module_id, relative_url)
  end

  def get_pdf_result(pdf_template_id, date_column = nil, start_date = nil, end_date = nil, sync = false, run_params = nil)
    count = count_rows(date_column, start_date, end_date)
    page_size = 5000
    fetch_limit = (count.to_f / page_size.to_f).ceil
    urls = []
    (0..fetch_limit).each do |i|
      urls.push(run_url(date_column: date_column, start_date: start_date, end_date: end_date, page_number: i, run_params: run_params, page_size: page_size, print_view: true, is_last_page: (i == fetch_limit)))
    end
    RxModule::Communicator::PDF.send(sync ? :generate_blocking : :generate, pdf_template_id, urls.map { |u| {url: u, render_selector: '#reportRendered'} }, name: name, notes: desc, module_id: report_target.module_id, target_id: report_target.report_id, action_id: 'run')
  end

  def count_rows(date_column = nil, start_date = nil, end_date = nil)
    resp = report_target.run(params_with_date_filter(date_column, start_date, end_date).merge(limit: 0))
    resp[:total_records]
  end
end
