# == Schema Information
#
# Table name: report_component_results
#
#  id                 :bigint           not null, primary key
#  data_format        :string
#  desc               :string
#  generating         :boolean
#  generation_error   :string
#  generation_percent :float
#  name               :string
#  parameters         :string
#  created_at         :datetime         not null
#  updated_at         :datetime         not null
#  report_target_id   :bigint
#
# Indexes
#
#  index_report_component_results_on_report_target_id  (report_target_id)
#
# Foreign Keys
#
#  fk_rails_...  (report_target_id => report_targets.id) ON DELETE => nullify ON UPDATE => cascade
#
require 'csv'

class ReportComponentResult < ApplicationRecord
  has_one_attached :data_file
  belongs_to :report_target

  def self.display_name
    "Report Component Result"
  end

  def download_link
    if data_file.attached? || (self.generating == false && self.generation_percent == 100)
      "/api/report_component_results/#{self.id}/download"
    else
      nil
    end
  end

  def data_file_disk_path
    if data_file.attached?
      ActiveStorage::Blob.service.send(:path_for, data_file.key)
    else
      nil
    end
  end

  def content_type
    case self.data_format
    when 'csv'
      'text/csv'
    when 'json'
      'application/json'
    else
      'text/plain'
    end
  end

  def content_extension
    case self.data_format
    when 'csv'
      'csv'
    when 'json'
      'json'
    else
      'txt'
    end
  end

  def filename
    "#{self.name || 'result'}.#{self.content_extension}"
  end

  def progress(completed_records, total_records)
    if total_records && total_records > 0 && completed_records
      percent = (completed_records / total_records) * 100
      Rails.logger.debug { "#{completed_records} / #{total_records} = #{percent}%)" }
      self.generation_percent = percent
      self.save
    end
  end

  def generate_file
    begin
      Rails.logger.info { {msg: "Making export", attributes: self.as_json} }
      self.data_file.purge
      self.generating = true
      self.generation_error = nil
      self.generation_percent = nil
      self.save
      case self.data_format
      when 'json'
        make_json
      when 'csv'
        make_csv
      else
        Rails.logger.warn { {msg: "Invalid export format", format: self.data_format, target: target, attributes: self.as_json} }
      end
      self.generation_error = nil
      self.generation_percent = 100
    rescue StandardError => err
      Rails.logger.error("Report result generation failed")
      self.generation_error = err.message
      self.generation_percent = nil
    ensure
      unless abort?
        self.generating = false
        self.save
      end
    end
  end

  def make_json
    #todo account for column name mappings
    parameters = JSON.parse(self.parameters)
    build_result do |file|
      file.write("[\n")
      first_record = true
      accumulated = 0.to_f
      total_records = 0.to_f
      report_target.paged_run(parameters) do |resp|
        resp[:hits].each do |record|
          record_str = record.to_json
          Rails.logger.debug { record_str }
          file.write(",\n") unless first_record
          first_record = false
          file.write(record_str)
        end
        total_records = resp[:total_records] if resp[:total_records]
        accumulated += resp[:hits].length
        progress(accumulated, total_records)
        abort?
      end
      file.write("]")
    end
  end

  def make_csv
    parameters = JSON.parse(self.parameters, symbolize_names: true)
    build_result do |file|
      csv = CSV.new(file)
      first_record = true
      accumulated = 0.to_f
      total_records = 0.to_f
      headers = []
      report_target.paged_run(parameters) do |resp|
        if first_record
          headers = parameters[:columns].map do |c|
            resolved_path = (resp[:column_mappings][c[:accessor].to_sym])
            resolved_path = (resolved_path[c[:aggregation]&.to_sym] || resolved_path[''.to_sym]) if resolved_path.is_a?(Hash)
            resolved_path = c[:accessor] if resolved_path.nil? || resolved_path == ''
            {
                path: resolved_path.split('.').map(&:to_sym),
                label: (c[:rename] && c[:rename] != '') ? c[:rename] : (report_target.display_name(c[:accessor]) || c[:accessor])
            }
          end
          csv << headers.map { |p| p[:label] }
          first_record = false
        end
        resp[:hits].each do |record|
          Rails.logger.debug { record }
          csv << headers.map do |h|
            obj = record
            h[:path].each do |step|
              if obj.is_a?(Array)
                obj = obj.map { |o| o&.[](step) }
              else
                obj = obj&.[](step)
              end
            end
            (obj.is_a?(Array) ? obj.map { |o| o.to_s }.join(',') : obj.to_s).gsub('&|>', ' ')
          end
        end
        total_records = resp[:total_records] if resp[:total_records]
        accumulated += resp[:hits].length
        progress(accumulated, total_records)
        abort?
      end
    end
  end

  def build_result
    Tempfile.create(["report_result", content_extension]) do |f|
      yield f
      f.flush
      f.rewind
      unless abort?
        self.data_file.attach(io: f, filename: File.basename(f.path), content_type: content_type, identify: false)
      end
    end
  end

  def abort?
    self.destroyed? || !ReportComponentResult.exists?(id: self.id)
  end

  def as_json
    super.merge(download_link: self.download_link)
  end
end
